package fr.gouv.beta.selexpert.controller;

import fr.gouv.beta.selexpert.dto.BuildExperts;
import fr.gouv.beta.selexpert.dto.CourtUser;
import fr.gouv.beta.selexpert.dto.DeleteExpert;
import fr.gouv.beta.selexpert.dto.ExpertCreationForm;
import fr.gouv.beta.selexpert.dto.PageOf;
import fr.gouv.beta.selexpert.model.Role;
import fr.gouv.beta.selexpert.model.UserAdmin;
import fr.gouv.beta.selexpert.model.UserClerk;
import fr.gouv.beta.selexpert.model.UserMagistrate;
import fr.gouv.beta.selexpert.repository.RequestRepository;
import fr.gouv.beta.selexpert.repository.UserRepository;
import fr.gouv.beta.selexpert.security.UserFromJwt;
import fr.gouv.beta.selexpert.service.AdminService;
import fr.gouv.beta.selexpert.service.ImportService;
import jakarta.validation.Valid;
import jakarta.validation.constraints.Min;
import jakarta.validation.constraints.NotNull;
import java.io.IOException;
import java.util.List;
import java.util.Objects;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.PageRequest;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.PostMapping;
import org.springframework.web.bind.annotation.PutMapping;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestParam;
import org.springframework.web.bind.annotation.RequestPart;
import org.springframework.web.bind.annotation.RestController;
import org.springframework.web.multipart.MultipartFile;

@RestController
@RequestMapping("/api/admin")
@Slf4j
@RequiredArgsConstructor
public class AdminController {

  private final AdminService adminService;
  private final ImportService importService;
  private final UserRepository userRepository;
  private final RequestRepository requestRepository;

  @Value("${selexpert.request-page-size}")
  private Integer requestPageSize;

  @Value("${selexpert.court-user-page-size}")
  private Integer courtUserPageSize;

  @PostMapping("/expert")
  public void createExpert(@RequestBody @Valid @NotNull ExpertCreationForm expertCreationForm) {
    adminService.createExpert(expertCreationForm);
  }

  @PutMapping("/delete-expert")
  public void deleteExpert(@RequestBody @Valid @NotNull DeleteExpert deleteExpert) {
    adminService.deleteExpert(deleteExpert.getExpertId());
  }

  @GetMapping("/court-users")
  public PageOf<CourtUser> getCourtUsers(
      @RequestParam("page") @Valid @Min(0) @NotNull Integer pageNumber,
      Integer appealCourtId,
      Integer jurisdictionId) {
    var pageRequest = PageRequest.of(pageNumber, courtUserPageSize);

    Page<Integer> courtUserIdPage; // Page<User> difficult with native query bc of inheritance
    if (Objects.nonNull(jurisdictionId))
      courtUserIdPage = userRepository.findCourtUserIdsOfJurisdiction(jurisdictionId, pageRequest);
    else if (Objects.nonNull(appealCourtId))
      courtUserIdPage = userRepository.findCourtUserIdsOfAppealCourt(appealCourtId, pageRequest);
    else courtUserIdPage = userRepository.findCourtUserIds(pageRequest);

    List<CourtUser> courtUsers =
        userRepository.findAllByIdInOrderByIdDesc(courtUserIdPage.stream().toList()).stream()
            .map(
                user -> {
                  CourtUser courtUser = CourtUser.from(user);
                  if (Objects.nonNull(courtUser)) {
                    courtUser.setRequestCount(
                        Role.MAGISTRATE.equals(user.getRole())
                            ? requestRepository.countByMagistrateAndClerkIsNull(
                                (UserMagistrate) user)
                            : requestRepository.countByClerk((UserClerk) user));
                  }
                  return courtUser;
                })
            .toList();
    return new PageOf<>(
        courtUsers,
        courtUserIdPage.getTotalPages(),
        courtUserIdPage.getTotalElements(),
        pageNumber,
        requestPageSize);
  }

  @GetMapping("/court-user-list")
  public List<CourtUser> getCourtUserList() {
    return userRepository.findAllByRoleInOrderByName(Role.MAGISTRATE, Role.CLERK).stream()
        .map(CourtUser::fromLight)
        .toList();
  }

  @PostMapping("/import/file")
  public String importExpertRawData(
      @RequestPart MultipartFile file, @NotNull Boolean dryrun, @UserFromJwt UserAdmin admin)
      throws IOException {
    return importService.importExpertRawData(file.getInputStream(), dryrun, admin);
  }

  @PostMapping("/import/build-experts")
  public String buildExperts(
      @RequestBody @Valid @NotNull BuildExperts buildExperts, @UserFromJwt UserAdmin admin) {
    return importService.buildExperts(buildExperts.getDryrun(), admin);
  }

  @GetMapping("/migration-auto")
  public String migrationAuto(@RequestParam("courtId") @Valid @NotNull Integer courtId) {
    return importService.migration(courtId, true);
  }

  @GetMapping("/migration-manual")
  public String migrationManual(@RequestParam("courtId") @Valid @NotNull Integer courtId) {
    return importService.migration(courtId, false);
  }

  @GetMapping("/update-experts")
  public String updateExperts(@RequestParam("courtId") @Valid @NotNull Integer courtId) {
    return importService.updateExperts(courtId);
  }

  @PostMapping("/import-temp")
  public String importTemp(@RequestPart MultipartFile file, @NotNull Boolean dryrun)
      throws IOException {
    return importService.importTemp(file.getInputStream(), dryrun);
  }
}
