package fr.gouv.beta.selexpert.controller;

import fr.gouv.beta.selexpert.dto.*;
import fr.gouv.beta.selexpert.model.User;
import fr.gouv.beta.selexpert.security.UserFromJwt;
import fr.gouv.beta.selexpert.service.AuthService;
import fr.gouv.beta.selexpert.service.JwtService;
import jakarta.validation.Valid;
import jakarta.validation.constraints.NotNull;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.springframework.http.ResponseEntity;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.security.authentication.AuthenticationManager;
import org.springframework.security.authentication.UsernamePasswordAuthenticationToken;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.AuthenticationException;
import org.springframework.web.bind.annotation.*;

@RestController
@RequestMapping("/api/auth")
@Slf4j
@RequiredArgsConstructor
public class AuthController {

  private final AuthService authService;
  private final AuthenticationManager authenticationManager;
  private final JwtService jwtService;

  @PostMapping("/login")
  public UserInfo login(@Valid @RequestBody LoginForm loginForm) {
    log.info("connection of {} ...", loginForm.getEmail());
    Authentication authentication;
    try {
      authentication =
          authenticationManager.authenticate(
              new UsernamePasswordAuthenticationToken(
                  loginForm.getEmail(), loginForm.getPassword()));
      log.info("OK");
    } catch (AuthenticationException e) {
      log.info("KO");
      throw e;
    }
    User user = (User) authentication.getPrincipal();
    UserInfo userInfo = authService.getUserInfo(user);
    userInfo.setAccessToken(jwtService.generateToken(user));
    return userInfo;
  }

  @GetMapping("/user-info")
  public UserInfo getUserInfo(@UserFromJwt User user) {
    return authService.getUserInfo(user);
  }

  @PutMapping("/user-info/court")
  @PreAuthorize("hasAnyRole('CLERK','MAGISTRATE')")
  public void updateCourtAccount(
      @RequestBody @Valid @NotNull CourtUserAccountUpdate courtUserAccountUpdate,
      @UserFromJwt User user) {
    authService.updateCourtAccount(courtUserAccountUpdate, user);
  }

  @PostMapping("/expert-activation")
  public ResponseEntity<Void> expertActivation(@RequestBody @Valid @NotNull EmailForm emailForm) {
    authService.activateExpert(emailForm.getEmail());
    return ResponseEntity.ok().build();
  }

  @PostMapping("/expert-activation/{token}")
  public ResponseEntity<Void> expertActivation(
      @RequestBody @Valid @NotNull PasswordForm passwordForm, @PathVariable String token) {
    authService.activateExpert(passwordForm.getPassword(), token);
    return ResponseEntity.ok().build();
  }

  @PostMapping("/court-registration")
  public ResponseEntity<Void> courtRegistration(
      @RequestBody @Valid @NotNull CourtRegistrationForm registrationForm) {
    authService.registerCourtUser(registrationForm);
    return ResponseEntity.ok().build();
  }

  @PostMapping("/password-reset")
  public ResponseEntity<Void> passwordReset(@RequestBody @Valid @NotNull EmailForm emailForm) {
    authService.passwordReset(emailForm.getEmail());
    return ResponseEntity.ok().build();
  }

  @PostMapping("/password-reset/{token}")
  public ResponseEntity<Void> passwordReset(
      @RequestBody @Valid @NotNull PasswordForm passwordForm, @PathVariable String token) {
    authService.passwordReset(passwordForm.getPassword(), token);
    return ResponseEntity.ok().build();
  }

  @GetMapping("/expert-activation-token-check/{token}")
  public ResponseEntity<Void> activationTokenCheck(@PathVariable String token) {
    authService.validateAuthTokenActivation(token);
    return ResponseEntity.ok().build();
  }

  @GetMapping("/password-reset-token-check/{token}")
  public ResponseEntity<Void> passwordResetTokenCheck(@PathVariable String token) {
    authService.validateAuthTokenPwdReset(token);
    return ResponseEntity.ok().build();
  }
}
