package fr.gouv.beta.selexpert.controller;

import fr.gouv.beta.selexpert.AppException;
import fr.gouv.beta.selexpert.dto.PageOf;
import fr.gouv.beta.selexpert.dto.RequestDto;
import fr.gouv.beta.selexpert.dto.RequestFilters;
import fr.gouv.beta.selexpert.dto.UserInfo;
import fr.gouv.beta.selexpert.model.AppealCourt;
import fr.gouv.beta.selexpert.model.CourtService;
import fr.gouv.beta.selexpert.model.ExpertiseNew;
import fr.gouv.beta.selexpert.model.Jurisdiction;
import fr.gouv.beta.selexpert.model.Request;
import fr.gouv.beta.selexpert.model.User;
import fr.gouv.beta.selexpert.model.UserClerk;
import fr.gouv.beta.selexpert.model.UserExpert;
import fr.gouv.beta.selexpert.model.UserMagistrate;
import fr.gouv.beta.selexpert.repository.AppealCourtRepository;
import fr.gouv.beta.selexpert.repository.CourtServiceRepository;
import fr.gouv.beta.selexpert.repository.ExpertiseNewRepository;
import fr.gouv.beta.selexpert.repository.JurisdictionRepository;
import fr.gouv.beta.selexpert.repository.RequestRepository;
import fr.gouv.beta.selexpert.repository.RequestSpecifications;
import fr.gouv.beta.selexpert.security.UserFromJwt;
import fr.gouv.beta.selexpert.service.CommonService;
import jakarta.validation.Valid;
import jakarta.validation.constraints.Min;
import jakarta.validation.constraints.NotNull;
import java.util.Comparator;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;
import java.util.stream.Collectors;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.PageRequest;
import org.springframework.data.domain.Sort;
import org.springframework.data.domain.Sort.Direction;
import org.springframework.http.HttpStatus;
import org.springframework.http.ResponseEntity;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.PutMapping;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestParam;
import org.springframework.web.bind.annotation.RestController;
import org.springframework.web.servlet.mvc.method.annotation.StreamingResponseBody;

@RestController
@RequestMapping("/api/common")
@Slf4j
@RequiredArgsConstructor
public class CommonController {

  private final AppealCourtRepository appealCourtRepository;
  private final JurisdictionRepository jurisdictionRepository;
  private final ExpertiseNewRepository expertiseNewRepository;
  private final CommonService commonService;
  private final RequestRepository requestRepository;
  private final CourtServiceRepository courtServiceRepository;

  @Value("${selexpert.production}")
  boolean isProduction;

  @Value("${selexpert.request-page-size}")
  private Integer requestPageSize;

  @Value("${selexpert.weekdays-to-reply}")
  private Integer weekdaysToReply;

  @GetMapping("/request")
  @PreAuthorize("isAuthenticated()")
  public PageOf<RequestDto> getAllRequests(
      @RequestParam("page") @Valid @Min(0) @NotNull Integer pageNumber,
      @Valid @NotNull RequestFilters requestFilters,
      @UserFromJwt User user) {
    requestFilters =
        switch (user.getRole()) {
          case MAGISTRATE ->
              RequestFilters.builder()
                  .magistrateId(user.getId())
                  .jurisdictionId(((UserMagistrate) user).getJurisdiction().getId())
                  .procNumChunk(requestFilters.getProcNumChunk())
                  .build();
          case CLERK ->
              RequestFilters.builder()
                  .clerkId(user.getId())
                  .jurisdictionId(((UserClerk) user).getJurisdiction().getId())
                  .procNumChunk(requestFilters.getProcNumChunk())
                  .build();
          case ADMIN -> requestFilters;
          default -> throw new AppException("role not allowed");
        };
    var pageRequest =
        PageRequest.of(pageNumber, requestPageSize, Sort.by(Direction.DESC, "createdOn"));
    Page<Request> requestPage =
        requestRepository.findAll(RequestSpecifications.all(requestFilters), pageRequest);
    List<RequestDto> requestDtoList =
        requestPage.stream().map(request -> RequestDto.from(request, weekdaysToReply)).toList();
    return new PageOf<>(
        requestDtoList,
        requestPage.getTotalPages(),
        requestPage.getTotalElements(),
        pageNumber,
        requestPageSize);
  }

  @GetMapping("/appeal-court")
  public List<AppealCourt> getAppealCourtList() {
    return appealCourtRepository.findAll(Sort.by(Sort.Direction.ASC, "name"));
  }

  @GetMapping("/jurisdiction")
  public Map<String, List<Jurisdiction>> getJurisdictionList() {
    return new TreeMap<>(
        jurisdictionRepository.findAll().stream()
            .sorted(Comparator.comparing(Jurisdiction::getName))
            .collect(
                Collectors.groupingBy(jurisdiction -> jurisdiction.getAppealCourt().getName())));
  }

  @GetMapping("/expertise")
  public List<ExpertiseNew> getExpertiseList() {
    return expertiseNewRepository.findAll(Sort.by(Sort.Direction.ASC, "id"));
  }

  @GetMapping("/court-service")
  public List<CourtService> getCourtServiceList() {
    return courtServiceRepository.findAllByParentIsNullOrderByRank();
  }

  @GetMapping("/download")
  @PreAuthorize("isAuthenticated()")
  public ResponseEntity<StreamingResponseBody> downloadFile(
      @RequestParam String fileName, @UserFromJwt User user) {
    return new ResponseEntity<>(commonService.download(fileName, user), HttpStatus.OK);
  }

  @GetMapping("/request-page")
  @PreAuthorize("isAuthenticated()")
  public Integer getRequestPage(
      @RequestParam("requestId") @Valid @NotNull Integer requestId, @UserFromJwt User user) {
    return switch (user.getRole()) {
      case MAGISTRATE ->
          requestRepository.getRequestPageNumberForMagistrate(
              requestId, user.getId(), requestPageSize);
      case CLERK ->
          requestRepository.getRequestPageNumberForClerk(requestId, user.getId(), requestPageSize);
      case EXPERT ->
          requestRepository.getRequestPageNumberForExpert(
              requestId, ((UserExpert) user).getExpert().getId(), requestPageSize);
      case ADMIN -> requestRepository.getRequestPageNumberForAdmin(requestId, requestPageSize);
    };
  }

  @PutMapping("/notifications")
  @PreAuthorize("isAuthenticated()")
  public UserInfo toggleNotifications(@UserFromJwt User user) {
    return UserInfo.from(commonService.toggleNotifications(user.getId()));
  }

  @GetMapping("/stats/public")
  public String getStatsUrlPublic() {
    return commonService.getStatsUrlPublic();
  }

  @GetMapping("/stats/admin-static")
  @PreAuthorize("hasRole('ADMIN')")
  public String getStatsUrlAdminStatic() {
    return commonService.getStatsUrlAdminStatic();
  }

  @GetMapping("/stats/admin-dynamic")
  @PreAuthorize("hasRole('ADMIN')")
  public String getStatsUrlAdminDynamic(@RequestParam Map<String, String> filters) {
    return commonService.getStatsUrlAdminDynamic(filters);
  }

  @GetMapping("/stats/court")
  @PreAuthorize("isAuthenticated()")
  public String getStatsUrlCourt(@RequestParam Map<String, String> filters) {
    return commonService.getStatsUrlCourt(filters);
  }
}
