package fr.gouv.beta.selexpert.controller;

import fr.gouv.beta.selexpert.AppException;
import fr.gouv.beta.selexpert.dto.*;
import fr.gouv.beta.selexpert.model.Expert;
import fr.gouv.beta.selexpert.model.Request;
import fr.gouv.beta.selexpert.model.RequestExpertStatus;
import fr.gouv.beta.selexpert.model.Role;
import fr.gouv.beta.selexpert.model.User;
import fr.gouv.beta.selexpert.model.UserClerk;
import fr.gouv.beta.selexpert.model.UserMagistrate;
import fr.gouv.beta.selexpert.repository.ExpertRepository;
import fr.gouv.beta.selexpert.repository.ExpertSpecifications;
import fr.gouv.beta.selexpert.repository.RequestExpertRepository;
import fr.gouv.beta.selexpert.repository.RequestRepository;
import fr.gouv.beta.selexpert.repository.UserClerkRepository;
import fr.gouv.beta.selexpert.repository.UserMagistrateRepository;
import fr.gouv.beta.selexpert.security.UserFromJwt;
import fr.gouv.beta.selexpert.service.CourtService;
import jakarta.validation.Valid;
import jakarta.validation.constraints.Min;
import jakarta.validation.constraints.NotNull;
import java.util.Comparator;
import java.util.List;
import java.util.Objects;
import java.util.stream.Collectors;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.cache.annotation.Cacheable;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.PageRequest;
import org.springframework.data.domain.Sort;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.web.bind.annotation.*;
import org.springframework.web.multipart.MultipartFile;

@RestController
@RequestMapping("/api/court")
@Slf4j
@RequiredArgsConstructor
public class CourtController {

  private final ExpertRepository expertRepository;
  private final UserClerkRepository userClerkRepository;
  private final UserMagistrateRepository userMagistrateRepository;
  private final RequestRepository requestRepository;
  private final RequestExpertRepository requestExpertRepository;
  private final CourtService courtService;

  @Value("${selexpert.weekdays-to-reply}")
  private Integer weekdaysToReply;

  @Value("${selexpert.expert-page-size}")
  private Integer expertPageSize;

  @GetMapping("/experts")
  public PageOf<ExpertDto> getExpertList(
      @RequestParam("page") @Valid @Min(0) Integer pageNumber,
      @Valid @NotNull ExpertFilters expertFilters) { // TODO sort by last name + general sorting
    expertFilters.setIncludeDeleted(false); // always override for now
    Page<Expert> expertPage =
        expertRepository.findAll(
            ExpertSpecifications.all(expertFilters),
            PageRequest.of(pageNumber, expertPageSize, Sort.by(Sort.Direction.ASC, "name")));
    List<ExpertDto> expertDtoList =
        expertPage.stream()
            .map(
                expert -> {
                  ExpertDto expertDto = ExpertDto.from(expert);
                  if (Objects.nonNull(expert.getUserExpert())) {
                    expertDto.setAcceptedRequestCount(
                        requestExpertRepository.countByExpertAndStatus(
                            expert, RequestExpertStatus.ACCEPTED));
                  }
                  return expertDto;
                })
            .collect(Collectors.toList());
    return new PageOf<>(
        expertDtoList,
        expertPage.getTotalPages(),
        expertPage.getTotalElements(),
        pageNumber,
        expertPageSize);
  }

  @GetMapping("/clerks")
  public List<ClerkDto> getClerkList(@UserFromJwt User user) {
    var jurisdiction =
        Role.MAGISTRATE.equals(user.getRole())
            ? ((UserMagistrate) user).getJurisdiction()
            : ((UserClerk) user).getJurisdiction();
    return userClerkRepository.findByJurisdiction(jurisdiction).stream()
        .map(clerk -> ClerkDto.from(clerk, true))
        .sorted(Comparator.comparing(ClerkDto::getName))
        .collect(Collectors.toList());
  }

  @GetMapping("/magistrates")
  @PreAuthorize("hasRole('CLERK')")
  public List<MagistrateDto> getMagistrateList(@UserFromJwt UserClerk userClerk) {
    return userMagistrateRepository.findByJurisdiction(userClerk.getJurisdiction()).stream()
        .map(MagistrateDto::from)
        .sorted(Comparator.comparing(MagistrateDto::getName))
        .collect(Collectors.toList());
  }

  @PostMapping("/request")
  public Integer createRequest(
      @RequestBody @Valid @NotNull RequestCreation requestCreation, @UserFromJwt User user) {
    if (Role.MAGISTRATE.equals(user.getRole())) {
      Request createdRequest =
          courtService.createRequest(requestCreation, (UserMagistrate) user, null);
      return createdRequest.getId();
    } else if (Role.CLERK.equals(user.getRole())) {
      UserClerk userClerk = (UserClerk) user;
      UserMagistrate userMagistrate =
          userMagistrateRepository
              .findByIdAndJurisdiction(
                  requestCreation.getMagistrateId(), userClerk.getJurisdiction())
              .orElseThrow(() -> new AppException("magistrate not found in clerk's jurisdiction"));
      Request createdRequest =
          courtService.createRequest(requestCreation, userMagistrate, userClerk);
      return createdRequest.getId();
    }
    throw new AppException("can't create request");
  }

  @GetMapping("/request/{requestId}")
  public RequestDto getRequest(@PathVariable Integer requestId, @UserFromJwt User user) {
    return RequestDto.from(courtService.getRequest(requestId, user), weekdaysToReply);
  }

  @PutMapping("/request/add-experts")
  public void addExpertsToRequest(
      @RequestBody @Valid @NotNull AddExpertsToRequest addExpertsToRequest,
      @UserFromJwt User user) {
    courtService.addExpertsToRequest(addExpertsToRequest, user);
  }

  @PutMapping("/request/cancel")
  public RequestDto cancelRequest(
      @RequestBody @Valid @NotNull CancelRequest cancelRequest, @UserFromJwt User user) {
    Request requestCancelled = courtService.cancelRequest(cancelRequest, user);
    return RequestDto.from(requestCancelled, weekdaysToReply);
  }

  @PostMapping("/request/file")
  public String addFile(
      @RequestPart @Valid @NotNull AddFile addFile,
      @RequestPart @NotNull MultipartFile file,
      @UserFromJwt User user) {
    return courtService.addFile(addFile, file, user);
  }

  @DeleteMapping("/request/file")
  public void deleteFile(
      @RequestBody @Valid @NotNull DeleteFile deleteFile, @UserFromJwt User user) {
    courtService.deleteFile(deleteFile, user);
  }

  @PostMapping("/request/note")
  public NoteDto addNote(@RequestBody @Valid @NotNull AddNote addNote, @UserFromJwt User user) {
    return NoteDto.from(courtService.addNote(addNote, user));
  }

  @DeleteMapping("/request/note")
  public void deleteNote(
      @RequestBody @Valid @NotNull DeleteNote deleteNote, @UserFromJwt User user) {
    courtService.deleteNote(deleteNote, user);
  }

  @PutMapping("/request/extra-info")
  public String updateExtraInfo(
      @RequestBody @Valid @NotNull ExtraInfoUpdate extraInfoUpdate, @UserFromJwt User user) {
    return courtService.updateExtraInfo(extraInfoUpdate, user);
  }

  @PutMapping("/request/deliberation-date")
  public void updateDeliberationDate(
      @RequestBody @Valid @NotNull DeliberationDateUpdate deliberationDateUpdate,
      @UserFromJwt User user) {
    courtService.updateDeliberationDate(deliberationDateUpdate, user);
  }

  @PutMapping("/request/clerks")
  public void updateClerks(
      @RequestBody @Valid @NotNull UpdateRequestClerks updateRequestClerks,
      @UserFromJwt User user) {
    courtService.updateClerks(updateRequestClerks, user);
  }

  @Cacheable("avg-acceptation-time")
  @GetMapping("/avg-acceptation-time")
  public Integer getAvgAcceptationTime() {
    return requestRepository.getAvgAcceptationTime();
  }
}
