package fr.gouv.beta.selexpert.controller;

import fr.gouv.beta.selexpert.AppException;
import fr.gouv.beta.selexpert.dto.AddNote;
import fr.gouv.beta.selexpert.dto.ExpertAvailabilityForm;
import fr.gouv.beta.selexpert.dto.ExpertUpdateData;
import fr.gouv.beta.selexpert.dto.NoteDto;
import fr.gouv.beta.selexpert.dto.PageOf;
import fr.gouv.beta.selexpert.dto.RequestDtoForExpert;
import fr.gouv.beta.selexpert.dto.RequestRefusalForm;
import fr.gouv.beta.selexpert.model.*;
import fr.gouv.beta.selexpert.repository.RequestRepository;
import fr.gouv.beta.selexpert.security.UserFromJwt;
import fr.gouv.beta.selexpert.service.ExpertService;
import jakarta.validation.Valid;
import jakarta.validation.constraints.Min;
import jakarta.validation.constraints.NotNull;
import java.util.List;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.PageRequest;
import org.springframework.data.domain.Sort;
import org.springframework.data.domain.Sort.Direction;
import org.springframework.http.ResponseEntity;
import org.springframework.web.bind.annotation.*;

@RestController
@RequestMapping("/api/expert")
@Slf4j
@RequiredArgsConstructor
public class ExpertController {

  private final RequestRepository requestRepository;
  private final ExpertService expertService;

  @Value("${selexpert.weekdays-to-reply}")
  private Integer weekdaysToReply;

  @Value("${selexpert.request-page-size}")
  private Integer requestPageSize;

  @GetMapping("/request")
  public PageOf<RequestDtoForExpert> getRequestList(
      @RequestParam("page") @Valid @Min(0) Integer pageNumber, @UserFromJwt UserExpert user) {
    Expert expert = user.getExpert();
    // display only the requests that have been sent to the expert
    var pageRequest =
        PageRequest.of(pageNumber, requestPageSize, Sort.by(Direction.DESC, "createdOn"));
    Page<Request> requestPage =
        requestRepository.findByRequestExpertsExpertAndRequestExpertsSentOnNotNull(
            expert, pageRequest);
    List<RequestDtoForExpert> requestDtoForExpertList =
        requestPage.stream()
            .map(
                request -> {
                  RequestExpert requestExpert =
                      request.getRequestExperts().stream()
                          .filter(re -> re.getExpert().getId().equals(expert.getId()))
                          .findFirst()
                          .orElseThrow();
                  return RequestDtoForExpert.from(requestExpert, weekdaysToReply);
                })
            .toList();
    return new PageOf<>(
        requestDtoForExpertList,
        requestPage.getTotalPages(),
        requestPage.getTotalElements(),
        pageNumber,
        requestPageSize);
  }

  @PostMapping("/request/{requestId}/accept")
  public ResponseEntity<Void> acceptRequest(
      @PathVariable Integer requestId, @UserFromJwt UserExpert user) {
    try {
      expertService.acceptRequest(requestId, user.getExpert().getId());
      return ResponseEntity.ok().build();
    } catch (AppException e) {
      return ResponseEntity.badRequest().build();
    }
  }

  @PostMapping("/request/{requestId}/refuse")
  public ResponseEntity<Void> refuseRequest(
      @PathVariable Integer requestId,
      @RequestBody @Valid @NotNull RequestRefusalForm requestRefusalForm,
      @UserFromJwt UserExpert user) {
    try {
      expertService.refuseRequest(requestId, user.getExpert().getId(), requestRefusalForm);
      return ResponseEntity.ok().build();
    } catch (AppException e) {
      return ResponseEntity.badRequest().build();
    }
  }

  @PostMapping("/request/question")
  public NoteDto addQuestion(
      @RequestBody @Valid @NotNull AddNote addNote, @UserFromJwt UserExpert user) {
    return NoteDto.from(expertService.addQuestion(addNote, user));
  }

  @PostMapping("/availability")
  public ResponseEntity<Void> updateAvailability(
      @RequestBody @Valid @NotNull ExpertAvailabilityForm expertAvailabilityForm,
      @UserFromJwt UserExpert user) {
    // TODO
    // for now keep unavailableFrom, maybe remove later
    try {
      //      Optional.ofNullable(user).map(userExpert -> userExpert).
      expertService.updateAvailability(user.getId(), expertAvailabilityForm);
      return ResponseEntity.ok().build();
    } catch (AppException e) {
      return ResponseEntity.badRequest().build();
    }
  }

  @PutMapping("")
  public String updateExpertData(
      @RequestBody @Valid @NotNull ExpertUpdateData expertUpdateData, @UserFromJwt User user) {
    Integer expertId;
    if (Role.EXPERT.equals(user.getRole())) {
      if (!expertUpdateData.getField().isUpdatableByExpert()) {
        throw new AppException("expert can't edit this data");
      }
      expertId = ((UserExpert) user).getExpert().getId();
    } else {
      expertId = expertUpdateData.getExpertId();
    }
    return expertService.updateExpertData(
        expertId, expertUpdateData.getField(), expertUpdateData.getValue());
  }
}
