package fr.gouv.beta.selexpert.service;

import fr.gouv.beta.selexpert.AppException;
import fr.gouv.beta.selexpert.dto.CourtRegistrationForm;
import fr.gouv.beta.selexpert.dto.CourtUserAccountUpdate;
import fr.gouv.beta.selexpert.dto.CourtUserServiceUpdate;
import fr.gouv.beta.selexpert.dto.UserInfo;
import fr.gouv.beta.selexpert.event.CourtUserActivatedEvent;
import fr.gouv.beta.selexpert.event.ExpertActivatedEvent;
import fr.gouv.beta.selexpert.event.PwdResetRequestedEvent;
import fr.gouv.beta.selexpert.model.AuthActivity;
import fr.gouv.beta.selexpert.model.AuthToken;
import fr.gouv.beta.selexpert.model.AuthTokenActivation;
import fr.gouv.beta.selexpert.model.AuthTokenPwdReset;
import fr.gouv.beta.selexpert.model.CourtService;
import fr.gouv.beta.selexpert.model.CourtUserCourtService;
import fr.gouv.beta.selexpert.model.Expert;
import fr.gouv.beta.selexpert.model.Jurisdiction;
import fr.gouv.beta.selexpert.model.Role;
import fr.gouv.beta.selexpert.model.User;
import fr.gouv.beta.selexpert.model.UserClerk;
import fr.gouv.beta.selexpert.model.UserMagistrate;
import fr.gouv.beta.selexpert.repository.AuthActivityRepository;
import fr.gouv.beta.selexpert.repository.AuthTokenActivationRepository;
import fr.gouv.beta.selexpert.repository.AuthTokenPwdResetRepository;
import fr.gouv.beta.selexpert.repository.CourtServiceRepository;
import fr.gouv.beta.selexpert.repository.ExpertRepository;
import fr.gouv.beta.selexpert.repository.JurisdictionRepository;
import fr.gouv.beta.selexpert.repository.UserRepository;
import jakarta.transaction.Transactional;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.ZoneId;
import java.util.Objects;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.context.ApplicationEventPublisher;
import org.springframework.scheduling.annotation.Scheduled;
import org.springframework.stereotype.Service;

@Service
@RequiredArgsConstructor
@Slf4j
public class AuthService {

  private final UserRepository userRepository;
  private final ExpertRepository expertRepository;
  private final JurisdictionRepository jurisdictionRepository;
  private final CourtServiceRepository courtServiceRepository;
  private final AuthTokenActivationRepository authTokenActivationRepository;
  private final AuthTokenPwdResetRepository authTokenPwdResetRepository;
  private final AuthActivityRepository authActivityRepository;
  private final UserService userService;
  private final ApplicationEventPublisher eventPublisher;

  @Value("${selexpert.auth-token-expiration}")
  private Integer authTokenExpiration;

  @Transactional
  public UserInfo getUserInfo(User user) {
    if (Objects.isNull(user)) return null;
    var now = LocalDate.now(ZoneId.of("Europe/Paris"));
    if (Objects.isNull(user.getLastActive()) || now.isAfter(user.getLastActive())) {
      user.setLastActive(now);
      if (!Role.ADMIN.equals(user.getRole())) {
        authActivityRepository
            .findById(new AuthActivity.AuthActivityId(user.getRole(), now))
            .ifPresentOrElse(
                authActivity -> authActivity.setCount(authActivity.getCount() + 1),
                () -> authActivityRepository.save(new AuthActivity(user.getRole(), now, 1)));
      }
    }
    return UserInfo.from(user);
  }

  @Transactional
  public void activateExpert(String email) {
    Expert expert = expertRepository.findByEmail(email).orElse(null);
    if (Objects.isNull(expert)) {
      throw new AppException(
          email + " can't be activated: not found", "ACTIVATION_EXPERT_NOT_FOUND");
    } else if (Objects.nonNull(expert.getUserExpert())) {
      throw new AppException(
          email + " can't be activated: already activated", "ACTIVATION_EXPERT_ACTIVATED");
    }
    authTokenActivationRepository
        .findByExpert(expert)
        .ifPresent(
            existingToken -> {
              authTokenActivationRepository.delete(existingToken);
              authTokenActivationRepository.flush(); // hibernate executes inserts before deletes
            });
    AuthTokenActivation authTokenActivation = new AuthTokenActivation(expert, authTokenExpiration);
    authTokenActivationRepository.save(authTokenActivation);
    eventPublisher.publishEvent(new ExpertActivatedEvent(expert, authTokenActivation.getToken()));
  }

  @Transactional
  public void activateExpert(String password, String token) {
    AuthTokenActivation authTokenActivation = validateAuthTokenActivation(token);
    userService.createExpertUser(authTokenActivation.getExpert(), password);
    authTokenActivationRepository.delete(authTokenActivation);
  }

  @Transactional
  public void registerCourtUser(CourtRegistrationForm registrationForm) {
    var email = registrationForm.getEmail();
    if (userRepository.existsByEmail(email)) {
      throw new AppException("email already exists");
    }
    User user =
        switch (registrationForm.getRole()) {
          case MAGISTRATE -> {
            UserMagistrate userMagistrate =
                UserMagistrate.builder().role(registrationForm.getRole()).build();
            updateCourtJurisdictionAndService(
                registrationForm.getJurisdictionId(),
                registrationForm.getService(),
                userMagistrate);
            yield userMagistrate;
          }
          case CLERK -> {
            UserClerk userClerk =
                UserClerk.builder()
                    .role(registrationForm.getRole())
                    .phoneLandline(registrationForm.getPhoneLandline())
                    .phoneMobile(registrationForm.getPhoneMobile())
                    .build();
            updateCourtJurisdictionAndService(
                registrationForm.getJurisdictionId(), registrationForm.getService(), userClerk);
            yield userClerk;
          }
          default -> throw new AppException("can't register for this role");
        };
    user.setEmail(email);
    user.setName(
        registrationForm.getLastName().toUpperCase()
            + " "
            + StringUtils.capitalize(registrationForm.getFirstName()));
    user.setSelfRegistration(true);
    userRepository.save(user);
    AuthTokenPwdReset authTokenPwdReset = new AuthTokenPwdReset(user, authTokenExpiration);
    authTokenPwdResetRepository.save(authTokenPwdReset);
    eventPublisher.publishEvent(new CourtUserActivatedEvent(user, authTokenPwdReset.getToken()));
  }

  @Transactional
  public void updateCourtAccount(CourtUserAccountUpdate courtUserAccountUpdate, User user) {
    updateCourtJurisdictionAndService(
        courtUserAccountUpdate.getJurisdictionId(), courtUserAccountUpdate.getService(), user);
  }

  @Transactional
  public void passwordReset(String email) {
    userRepository
        .findByEmail(email)
        .ifPresent(
            user -> {
              authTokenPwdResetRepository
                  .findByUser(user)
                  .ifPresent(
                      existingToken -> {
                        authTokenPwdResetRepository.delete(existingToken);
                        authTokenPwdResetRepository
                            .flush(); // hibernate executes inserts before deletes
                      });
              AuthTokenPwdReset authTokenPwdReset =
                  new AuthTokenPwdReset(user, authTokenExpiration);
              authTokenPwdResetRepository.save(authTokenPwdReset);
              eventPublisher.publishEvent(
                  new PwdResetRequestedEvent(user, authTokenPwdReset.getToken()));
            });
  }

  @Transactional
  public void passwordReset(String password, String token) {
    AuthTokenPwdReset authTokenPwdReset = validateAuthTokenPwdReset(token);
    userService.updateUserPassword(authTokenPwdReset.getUser(), password);
    authTokenPwdResetRepository.delete(authTokenPwdReset);
  }

  public AuthTokenActivation validateAuthTokenActivation(String token) {
    AuthTokenActivation authTokenActivation =
        authTokenActivationRepository.findByToken(token).orElse(null);
    checkAuthToken(authTokenActivation, token);
    return authTokenActivation;
  }

  public AuthTokenPwdReset validateAuthTokenPwdReset(String token) {
    AuthTokenPwdReset authTokenPwdReset =
        authTokenPwdResetRepository.findByToken(token).orElse(null);
    checkAuthToken(authTokenPwdReset, token);
    return authTokenPwdReset;
  }

  private void checkAuthToken(AuthToken authToken, String token) {
    if (Objects.isNull(authToken)) {
      throw new AppException("authToken not found " + token);
    } else if (authToken.getExpiryDate().isBefore(LocalDateTime.now())) {
      throw new AppException("authToken expired " + token);
    }
  }

  private void updateCourtJurisdictionAndService(
      Integer jurisdictionId, CourtUserServiceUpdate service, User user) {
    Jurisdiction jurisdiction =
        Objects.nonNull(jurisdictionId)
            ? jurisdictionRepository.getReferenceById(jurisdictionId)
            : null;
    CourtService courtService =
        courtServiceRepository
            .findById(service.getServiceId())
            .orElseThrow(() -> new AppException("court service not found"));
    if (courtService.isFreeInput() && StringUtils.isBlank(service.getServiceOther())) {
      throw new AppException("this court service needs free input");
    }
    switch (user.getRole()) {
      case MAGISTRATE -> {
        UserMagistrate userMagistrate = (UserMagistrate) user;
        if (Objects.nonNull(jurisdiction)) userMagistrate.setJurisdiction(jurisdiction);
        if (Objects.nonNull(userMagistrate.getService())
            && userMagistrate.getService().getCourtService().equals(courtService)) {
          userMagistrate.getService().setOther(service.getServiceOther());
        } else {
          userMagistrate.setService(
              new CourtUserCourtService(user, courtService, service.getServiceOther()));
        }
      }
      case CLERK -> {
        UserClerk userClerk = (UserClerk) user;
        if (Objects.nonNull(jurisdiction)) userClerk.setJurisdiction(jurisdiction);
        if (Objects.nonNull(userClerk.getService())
            && userClerk.getService().getCourtService().equals(courtService)) {
          userClerk.getService().setOther(service.getServiceOther());
        } else {
          userClerk.setService(
              new CourtUserCourtService(user, courtService, service.getServiceOther()));
        }
      }
    }
  }

  @Transactional
  @Scheduled(cron = "0 0 0 * * *", zone = "Europe/Paris")
  public void deleteExpiredAuthTokens() {
    log.info("deleteExpiredAuthTokens()");
    var now = LocalDateTime.now();
    authTokenActivationRepository.deleteByExpiryDateBefore(now);
    authTokenPwdResetRepository.deleteByExpiryDateBefore(now);
  }
}
