package fr.gouv.beta.selexpert.service;

import com.nimbusds.jose.JOSEException;
import com.nimbusds.jose.JWSAlgorithm;
import com.nimbusds.jose.JWSHeader;
import com.nimbusds.jose.JWSObject;
import com.nimbusds.jose.JWSSigner;
import com.nimbusds.jose.Payload;
import com.nimbusds.jose.crypto.MACSigner;
import fr.gouv.beta.selexpert.AppException;
import fr.gouv.beta.selexpert.event.RequestDeliberationReminderEvent;
import fr.gouv.beta.selexpert.model.File;
import fr.gouv.beta.selexpert.model.FileType;
import fr.gouv.beta.selexpert.model.Request;
import fr.gouv.beta.selexpert.model.RequestExpertStatus;
import fr.gouv.beta.selexpert.model.RequestStatus;
import fr.gouv.beta.selexpert.model.User;
import fr.gouv.beta.selexpert.repository.FileRepository;
import fr.gouv.beta.selexpert.repository.RequestRepository;
import fr.gouv.beta.selexpert.repository.UserRepository;
import jakarta.transaction.Transactional;
import java.time.Instant;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.ZoneId;
import java.util.HashMap;
import java.util.Map;
import java.util.Objects;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.context.ApplicationEventPublisher;
import org.springframework.scheduling.annotation.Scheduled;
import org.springframework.stereotype.Service;
import org.springframework.web.servlet.mvc.method.annotation.StreamingResponseBody;

@Service
@RequiredArgsConstructor
@Slf4j
public class CommonService {

  private final FileService fileService;
  private final CourtService courtService;
  private final RequestRepository requestRepository;
  private final FileRepository fileRepository;
  private final ApplicationEventPublisher eventPublisher;
  private final UserRepository userRepository;

  @Value("${selexpert.days-to-remind-deliberation}")
  private Integer daysToRemindDeliberation;

  @Value("${selexpert.months-to-archive-request}")
  private Integer monthsToArchiveRequest;

  @Value("${selexpert.metabase.url}")
  private String metabaseUrl;

  @Value("${selexpert.metabase.secret-key}")
  private String metabaseSecretKey;

  @Value("${selexpert.metabase.dashboard-id-public}")
  private Integer metabaseDashboardIdPublic;

  @Value("${selexpert.metabase.dashboard-id-admin-static}")
  private Integer metabaseDashboardIdAdminStatic;

  @Value("${selexpert.metabase.dashboard-id-admin-dynamic}")
  private Integer metabaseDashboardIdAdminDynamic;

  @Value("${selexpert.metabase.dashboard-id-court}")
  private Integer metabaseDashboardIdCourt;

  public StreamingResponseBody download(String fileName, User user) {
    File file =
        fileRepository
            .findByName(fileName)
            .orElseThrow(() -> new AppException("file " + fileName + " does not exist"));
    Request request = file.getRequest();
    boolean allowed =
        switch (user.getRole()) {
          case MAGISTRATE -> request.getMagistrate().getId().equals(user.getId());
          case CLERK ->
              request.getRequestClerks().stream().anyMatch(rq -> rq.getId().equals(user.getId()));
          case EXPERT ->
              request.getRequestExperts().stream()
                  .anyMatch(
                      re ->
                          Objects.nonNull(re.getExpert().getUserExpert())
                              && re.getExpert().getUserExpert().getId().equals(user.getId())
                              && (FileType.ATTACHMENT.equals(file.getType())
                                  || (FileType.DECISION.equals(file.getType())
                                      && RequestExpertStatus.ACCEPTED.equals(re.getStatus()))));
          case ADMIN -> false;
        };
    if (!allowed) {
      throw new AppException("file " + fileName + " can't be accessed by this user");
    }
    return fileService.download(fileName);
  }

  @Transactional
  @Scheduled(cron = "0 0 1 * * *", zone = "Europe/Paris")
  public void archiveRequests() {
    log.info("archiveRequests()");
    var thresholdDate = LocalDateTime.now().minusMonths(monthsToArchiveRequest);
    requestRepository
        .getRequestsToBeArchived(thresholdDate)
        .forEach(
            request -> {
              request.getFiles().forEach(file -> fileService.delete(file.getName()));
              request.setArchived(true);
            });
  }

  @Transactional
  @Scheduled(cron = "0 1 0 * * *", zone = "Europe/Paris")
  public void checkDeliberationDate() { // 1 minute after midnight
    log.info("checkDeliberationDate()");
    var today = LocalDate.now(ZoneId.of("Europe/Paris"));
    requestRepository
        .getRequestsToRemindDeliberation(today, daysToRemindDeliberation)
        .forEach(
            request -> eventPublisher.publishEvent(new RequestDeliberationReminderEvent(request)));
  }

  @Transactional
  @Scheduled(cron = "0 1 0 * * *", zone = "Europe/Paris")
  public void cancelDeliberatedRequests() { // 1 minute after midnight
    log.info("cancelDeliberatedRequests()");
    var today = LocalDate.now(ZoneId.of("Europe/Paris"));
    requestRepository
        .findByStatusAndDeliberationDateBefore(RequestStatus.OPEN, today)
        .forEach(request -> courtService.cancelRequest(request, true));
  }

  @Transactional
  public User toggleNotifications(Integer userId) {
    User user = userRepository.findById(userId).orElseThrow();
    user.setAllNotifications(Boolean.FALSE.equals(user.getAllNotifications()));
    return user;
  }

  public String getStatsUrlPublic() {
    return generateMetabaseUrl(metabaseDashboardIdPublic, Map.of());
  }

  public String getStatsUrlAdminStatic() {
    return generateMetabaseUrl(metabaseDashboardIdAdminStatic, Map.of());
  }

  public String getStatsUrlAdminDynamic(Map<String, String> filters) {
    return generateMetabaseUrl(metabaseDashboardIdAdminDynamic, filters);
  }

  public String getStatsUrlCourt(Map<String, String> filters) {
    return generateMetabaseUrl(metabaseDashboardIdCourt, filters);
  }

  private String generateMetabaseUrl(int dashboardId, Map<String, String> params) {
    try {
      JWSSigner signer = new MACSigner(metabaseSecretKey);
      Map<String, Object> payload = new HashMap<>();
      payload.put("resource", Map.of("dashboard", dashboardId));
      payload.put("params", params);
      payload.put("exp", Instant.now().getEpochSecond() + 600);
      JWSObject jwsObject = new JWSObject(new JWSHeader(JWSAlgorithm.HS256), new Payload(payload));
      jwsObject.sign(signer);
      return StringUtils.replace(metabaseUrl, "{}", jwsObject.serialize());
    } catch (JOSEException e) {
      log.error("metabase url error", e);
      return "";
    }
  }
}
