package fr.gouv.beta.selexpert.service;

import fr.gouv.beta.selexpert.AppException;
import fr.gouv.beta.selexpert.dto.AddExpertsToRequest;
import fr.gouv.beta.selexpert.dto.AddFile;
import fr.gouv.beta.selexpert.dto.AddNote;
import fr.gouv.beta.selexpert.dto.CancelRequest;
import fr.gouv.beta.selexpert.dto.DeleteFile;
import fr.gouv.beta.selexpert.dto.DeleteNote;
import fr.gouv.beta.selexpert.dto.DeliberationDateUpdate;
import fr.gouv.beta.selexpert.dto.ExtraInfoUpdate;
import fr.gouv.beta.selexpert.dto.RequestCreation;
import fr.gouv.beta.selexpert.dto.UpdateRequestClerks;
import fr.gouv.beta.selexpert.event.ClerksAddedEvent;
import fr.gouv.beta.selexpert.event.DecisionAddedEvent;
import fr.gouv.beta.selexpert.event.ExpertsAddedEvent;
import fr.gouv.beta.selexpert.event.QuestionAnsweredEvent;
import fr.gouv.beta.selexpert.event.RequestCancelledEvent;
import fr.gouv.beta.selexpert.event.RequestCreatedEvent;
import fr.gouv.beta.selexpert.event.RequestDeliberationReminderEvent;
import fr.gouv.beta.selexpert.model.*;
import fr.gouv.beta.selexpert.repository.ExpertRepository;
import fr.gouv.beta.selexpert.repository.FileRepository;
import fr.gouv.beta.selexpert.repository.NoteRepository;
import fr.gouv.beta.selexpert.repository.RequestRepository;
import fr.gouv.beta.selexpert.repository.UserClerkRepository;
import jakarta.transaction.Transactional;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.ZoneId;
import java.time.temporal.ChronoUnit;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.Objects;
import java.util.Optional;
import java.util.Set;
import java.util.stream.Collectors;
import java.util.stream.IntStream;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.context.ApplicationEventPublisher;
import org.springframework.stereotype.Service;
import org.springframework.web.multipart.MultipartFile;

@Service
@RequiredArgsConstructor
@Slf4j
public class CourtService {

  private final ExpertRepository expertRepository;
  private final UserClerkRepository userClerkRepository;
  private final RequestRepository requestRepository;
  private final ApplicationEventPublisher eventPublisher;
  private final FileService fileService;
  private final FileRepository fileRepository;
  private final NoteRepository noteRepository;

  @Value("${selexpert.days-to-remind-deliberation}")
  private Integer daysToRemindDeliberation;

  @Transactional
  public Request createRequest(
      RequestCreation requestCreation, UserMagistrate magistrate, UserClerk clerk) {
    Request request =
        Request.builder()
            .jurisdiction(magistrate.getJurisdiction())
            .magistrate(magistrate)
            .clerk(clerk)
            .procedureId(requestCreation.getProcedureId())
            .extraInfo(requestCreation.getExtraInfo())
            .urgent(requestCreation.getUrgent())
            .createdOn(LocalDateTime.now())
            .status(RequestStatus.OPEN)
            .deliberationDate(requestCreation.getDeliberationDate())
            .build();
    addExperts(requestCreation.getExpertIds(), request);
    addClerks(requestCreation.getClerkIds(), request, magistrate, clerk);
    Request savedRequest = requestRepository.save(request);
    addDuplicatedFiles(requestCreation.getDuplicatedFiles(), savedRequest, magistrate, clerk);
    eventPublisher.publishEvent(new RequestCreatedEvent(request));
    checkDeliberationDate(request);
    return savedRequest;
  }

  @Transactional
  public Request getRequest(Integer requestId, User user) {
    return getRequestIfInvolved(requestId, user);
  }

  @Transactional
  public void addExpertsToRequest(AddExpertsToRequest addExpertsToRequest, User user) {
    Request request = getRequestIfInvolved(addExpertsToRequest.getRequestId(), user);
    checkCanBeUpdated(request, RequestStatus.OPEN, RequestStatus.REFUSED);
    addExperts(addExpertsToRequest.getExpertIds(), request);
    request.setStatus(RequestStatus.OPEN);
    eventPublisher.publishEvent(new ExpertsAddedEvent(request));
  }

  @Transactional
  public String addFile(AddFile addFile, MultipartFile file, User user) {
    Request request = getRequestIfInvolved(addFile.getRequestId(), user);
    switch (addFile.getType()) {
      case ATTACHMENT -> {
        checkCanBeUpdated(request, RequestStatus.OPEN);
        // WARNING: will not work if parallel upload of multiple files
        var ranking =
            request.getFiles().stream()
                    .map(File::getRanking)
                    .filter(Objects::nonNull)
                    .max(Integer::compareTo)
                    .orElse(0)
                + 1;
        String fileName = request.getId() + "/attachment-" + ranking + ".pdf";
        fileService.upload(file, fileName);
        File addedFile =
            File.builder()
                .name(fileName)
                .description(addFile.getDescription())
                .request(request)
                .ranking(ranking)
                .type(addFile.getType())
                .build();
        fileRepository.save(addedFile);
        return addedFile.getName();
      }
      case DECISION -> {
        checkCanBeUpdated(request, RequestStatus.ACCEPTED);
        String fileName = request.getId() + "/decision.pdf";
        fileService.upload(file, fileName);
        File addedFile =
            File.builder().name(fileName).request(request).type(addFile.getType()).build();
        fileRepository.save(addedFile);
        request.setStatus(RequestStatus.COMPLETE);
        var expertWhoAccepted =
            request.getRequestExperts().stream()
                .filter(re -> RequestExpertStatus.ACCEPTED.equals(re.getStatus()))
                .findAny()
                .map(RequestExpert::getExpert)
                .orElseThrow();
        eventPublisher.publishEvent(new DecisionAddedEvent(request, expertWhoAccepted));
        return addedFile.getName();
      }
      default -> throw new AppException("unexpected filetype");
    }
  }

  @Transactional
  public void deleteFile(DeleteFile deleteFile, User user) {
    var requestId = Integer.valueOf(deleteFile.getFileName().split("/")[0]);
    Request request = getRequestIfInvolved(requestId, user);
    if (request.getFiles().removeIf(file -> file.getName().equals(deleteFile.getFileName()))) {
      fileService.delete(deleteFile.getFileName());
    }
  }

  @Transactional
  public Note addNote(AddNote addNote, User user) {
    Request request = getRequestIfInvolved(addNote.getRequestId(), user);
    boolean isAnswer = Objects.nonNull(addNote.getQuestionId());
    Note question = null;
    if (isAnswer) {
      question =
          request.getNotes().stream()
              .filter(
                  note ->
                      note.getId().equals(addNote.getQuestionId())
                          && note.getType().equals(NoteType.QUESTION))
              .findFirst()
              .orElseThrow(() -> new AppException("question not found in this request"));
    }
    Note note =
        Note.builder()
            .content(addNote.getContent())
            .request(request)
            .createdOn(LocalDateTime.now())
            .user(user)
            .type(isAnswer ? NoteType.ANSWER : NoteType.INTERNAL)
            .questionId(isAnswer ? addNote.getQuestionId() : null)
            .build();
    noteRepository.save(note);
    if (isAnswer) {
      eventPublisher.publishEvent(
          new QuestionAnsweredEvent(request, ((UserExpert) question.getUser()).getExpert()));
    }
    return note;
  }

  @Transactional
  public void deleteNote(DeleteNote deleteNote, User user) {
    noteRepository
        .findById(deleteNote.getNoteId())
        .ifPresent(
            note ->
                getRequestIfInvolved(note.getRequest().getId(), user)
                    .getNotes()
                    .removeIf(n -> n.getId().equals(deleteNote.getNoteId())));
  }

  @Transactional
  public String updateExtraInfo(ExtraInfoUpdate extraInfoUpdate, User user) {
    Request request = getRequestIfInvolved(extraInfoUpdate.getRequestId(), user);
    checkCanBeUpdated(request, RequestStatus.OPEN);
    request.setExtraInfo(extraInfoUpdate.getExtraInfo());
    return request.getExtraInfo();
  }

  @Transactional
  public void updateDeliberationDate(DeliberationDateUpdate deliberationDateUpdate, User user) {
    Request request = getRequestIfInvolved(deliberationDateUpdate.getRequestId(), user);
    checkCanBeUpdated(request, RequestStatus.OPEN);
    request.setDeliberationDate(deliberationDateUpdate.getDeliberationDate());
    checkDeliberationDate(request);
  }

  @Transactional
  public void updateClerks(UpdateRequestClerks updateRequestClerks, User user) {
    Request request = getRequestIfInvolved(updateRequestClerks.getRequestId(), user);
    checkCanBeUpdated(request, RequestStatus.OPEN);
    request
        .getRequestClerks()
        .removeIf(
            requestClerk -> !updateRequestClerks.getClerkIds().contains(requestClerk.getId()));
    List<UserClerk> clerksToAdd =
        updateRequestClerks.getClerkIds().stream()
            .filter(
                clerkId ->
                    !request.getRequestClerks().stream()
                        .map(User::getId)
                        .toList()
                        .contains(clerkId))
            .map(
                clerkId ->
                    userClerkRepository
                        .findByIdAndJurisdiction(clerkId, request.getMagistrate().getJurisdiction())
                        .orElse(null))
            .filter(Objects::nonNull)
            .toList();
    request.getRequestClerks().addAll(clerksToAdd);
    eventPublisher.publishEvent(new ClerksAddedEvent(request, clerksToAdd));
  }

  @Transactional
  public Request cancelRequest(CancelRequest cancelRequest, User user) {
    Request request = getRequestIfInvolved(cancelRequest.getRequestId(), user);
    cancelRequest(request, false);
    return request;
  }

  @Transactional
  protected void cancelRequest(Request request, boolean autoCancel) { // use internally only
    checkCanBeUpdated(request, RequestStatus.OPEN);
    request.getRequestExperts().stream()
        .filter(
            re ->
                Arrays.asList(
                        RequestExpertStatus.SENDING,
                        RequestExpertStatus.SENT,
                        RequestExpertStatus.QUEUED)
                    .contains(re.getStatus()))
        .forEach(re -> re.updateStatus(RequestExpertStatus.CANCELLED));
    request.setStatus(autoCancel ? RequestStatus.CANCELLED_AUTO : RequestStatus.CANCELLED);
    eventPublisher.publishEvent(new RequestCancelledEvent(request));
  }

  private void addExperts(List<Integer> expertIds, Request request) {
    if (CollectionUtils.isEmpty(expertIds)) {
      throw new AppException("no experts to add");
    }
    var currentExpertCount = request.getRequestExperts().size();
    request
        .getRequestExperts()
        .addAll(
            IntStream.range(0, expertIds.size())
                .mapToObj(
                    i -> {
                      Expert expert =
                          expertRepository
                              .findById(expertIds.get(i))
                              .orElseThrow(() -> new AppException("expert not found"));
                      if (StringUtils.isBlank(expert.getEmail())) {
                        throw new AppException("can't select this expert (no email)");
                      }
                      if (Optional.ofNullable(expert.getDeleted()).orElse(false)) {
                        throw new AppException("can't select this expert (deleted)");
                      }
                      if (Objects.nonNull(expert.getUserExpert())
                          && Objects.nonNull(expert.getUserExpert().getUnavailableFrom())) {
                        throw new AppException("can't select this expert (unavailable)");
                      }
                      var requestExpert =
                          new RequestExpert(request, expert, i + 1 + currentExpertCount);
                      if (request.isUrgent() // send to all
                          || (i == 0 // send to first of the list
                              && (currentExpertCount == 0 // request creation
                                  || RequestStatus.REFUSED.equals(request.getStatus())))) {
                        requestExpert.updateStatus(RequestExpertStatus.SENDING);
                      }
                      return requestExpert;
                    })
                .toList());
  }

  private void addClerks(
      List<Integer> clerkIds, Request request, UserMagistrate magistrate, UserClerk clerk) {
    Set<Integer> clerkIdSet = new HashSet<>();
    if (Objects.nonNull(clerk)) {
      clerkIdSet.add(clerk.getId());
    }
    if (CollectionUtils.isNotEmpty(clerkIds)) {
      clerkIdSet.addAll(clerkIds);
    }
    if (CollectionUtils.isNotEmpty(clerkIdSet)) {
      request.setRequestClerks(
          clerkIdSet.stream()
              .map(
                  clerkId ->
                      userClerkRepository
                          .findByIdAndJurisdiction(clerkId, magistrate.getJurisdiction())
                          .orElseThrow(
                              () ->
                                  new AppException("clerk not found in magistrate's jurisdiction")))
              .collect(Collectors.toList()));
    }
  }

  private Request getRequestIfInvolved(Integer requestId, User user) {
    return switch (user.getRole()) {
      case MAGISTRATE ->
          requestRepository
              .findByIdAndJurisdictionAndMagistrate(
                  requestId, ((UserMagistrate) user).getJurisdiction(), (UserMagistrate) user)
              .orElseThrow(() -> new AppException("magistrate not involved in this request"));
      case CLERK ->
          requestRepository
              .findByIdAndJurisdictionAndRequestClerks(
                  requestId, ((UserClerk) user).getJurisdiction(), (UserClerk) user)
              .orElseThrow(() -> new AppException("clerk not involved in this request"));
      case ADMIN ->
          requestRepository
              .findById(requestId)
              .orElseThrow(() -> new AppException("request not found"));
      default -> throw new AppException("not authorised to update this request");
    };
  }

  private void checkCanBeUpdated(Request request, RequestStatus... requestStatuses) {
    if (!Arrays.asList(requestStatuses).contains(request.getStatus()) || request.isArchived()) {
      throw new AppException(
          "can't update request that's not "
              + Arrays.toString(requestStatuses)
              + " or that's archived");
    }
  }

  private void checkDeliberationDate(Request request) {
    var today = LocalDate.now(ZoneId.of("Europe/Paris"));
    if (Objects.nonNull(request.getDeliberationDate())
        && ChronoUnit.DAYS.between(today, request.getDeliberationDate())
            <= daysToRemindDeliberation) {
      eventPublisher.publishEvent(new RequestDeliberationReminderEvent(request));
    }
  }

  private void addDuplicatedFiles(
      List<String> duplicatedFiles, Request request, UserMagistrate magistrate, UserClerk clerk) {
    if (CollectionUtils.isEmpty(duplicatedFiles)) return;
    Integer duplicatedRequestId = Integer.parseInt(duplicatedFiles.get(0).split("/")[0]);
    Request duplicatedRequest =
        getRequestIfInvolved(duplicatedRequestId, Objects.nonNull(magistrate) ? magistrate : clerk);
    checkCanBeDuplicated(duplicatedRequest, duplicatedFiles);
    int ranking = 0;
    for (String duplicatedFilePath : duplicatedFiles) {
      String fileName = request.getId() + "/attachment-" + ++ranking + ".pdf";
      if (!fileService.copy(duplicatedFilePath, fileName)) continue;
      File addedFile =
          File.builder()
              .name(fileName)
              .description(
                  duplicatedRequest.getFiles().stream()
                      .filter(
                          f ->
                              duplicatedFilePath.equals(f.getName())
                                  && FileType.ATTACHMENT.equals(f.getType()))
                      .map(File::getDescription)
                      .findFirst()
                      .orElse(""))
              .request(request)
              .ranking(ranking)
              .type(FileType.ATTACHMENT)
              .build();
      fileRepository.save(addedFile);
    }
  }

  private void checkCanBeDuplicated(Request duplicatedRequest, List<String> duplicatedFiles) {
    try {
      checkCanBeUpdated(duplicatedRequest, RequestStatus.CANCELLED, RequestStatus.CANCELLED_AUTO);
      if (duplicatedFiles.size()
          > duplicatedFiles.stream()
              .filter(f -> f.startsWith(duplicatedRequest.getId() + "/"))
              .toList()
              .size()) throw new AppException("different request ids in duplicated files");
    } catch (Exception e) {
      log.error("request duplication is not possible", e);
      throw e;
    }
  }
}
