package fr.gouv.beta.selexpert.service;

import static java.util.stream.Collectors.groupingBy;

import fr.gouv.beta.selexpert.AppException;
import fr.gouv.beta.selexpert.dto.AddNote;
import fr.gouv.beta.selexpert.dto.ExpertAvailabilityForm;
import fr.gouv.beta.selexpert.dto.ExpertUpdateData.UpdatableData;
import fr.gouv.beta.selexpert.dto.RequestRefusalForm;
import fr.gouv.beta.selexpert.event.ExpertIsNowAvailableEvent;
import fr.gouv.beta.selexpert.event.ExpertIsStillUnavailableEvent;
import fr.gouv.beta.selexpert.event.QuestionAskedEvent;
import fr.gouv.beta.selexpert.event.RequestAcceptedEvent;
import fr.gouv.beta.selexpert.event.RequestRefusedAutoEvent;
import fr.gouv.beta.selexpert.event.RequestRefusedEvent;
import fr.gouv.beta.selexpert.model.*;
import fr.gouv.beta.selexpert.repository.ExpertRepository;
import fr.gouv.beta.selexpert.repository.NoteRepository;
import fr.gouv.beta.selexpert.repository.RequestExpertRepository;
import fr.gouv.beta.selexpert.repository.UserExpertRepository;
import fr.gouv.beta.selexpert.util.DateUtils;
import fr.gouv.beta.selexpert.util.TextUtils;
import jakarta.transaction.Transactional;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.ZoneId;
import java.time.format.DateTimeFormatter;
import java.time.temporal.ChronoUnit;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;
import java.util.stream.Collectors;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.context.ApplicationEventPublisher;
import org.springframework.scheduling.annotation.Scheduled;
import org.springframework.stereotype.Service;

@Service
@RequiredArgsConstructor
@Slf4j
public class ExpertService {

  private final RequestExpertRepository requestExpertRepository;
  private final UserExpertRepository userExpertRepository;
  private final ExpertRepository expertRepository;
  private final NoteRepository noteRepository;
  private final ApplicationEventPublisher eventPublisher;

  @Value(
      "#{T(java.time.format.DateTimeFormatter).ofPattern('${selexpert.date-format.unavailable}')}")
  private DateTimeFormatter dtfUnavailable;

  @Value("${selexpert.weekdays-to-reply}")
  private Integer weekdaysToReply;

  @Transactional
  public void acceptRequest(Integer requestId, Integer expertId) {
    RequestExpert requestExpert =
        getRequestExpertIfInvolved(
            requestId, expertId, List.of(RequestExpertStatus.SENDING, RequestExpertStatus.SENT));
    Request request = requestExpert.getRequest();
    requestExpert.updateStatus(RequestExpertStatus.ACCEPTED);
    var otherRequestExperts =
        request.getRequestExperts().stream()
            .filter(re -> !re.getExpert().getId().equals(requestExpert.getExpert().getId()));
    otherRequestExperts // filter out REFUSED / CANCELLED
        .filter(
            re ->
                List.of(
                        RequestExpertStatus.QUEUED,
                        RequestExpertStatus.SENDING,
                        RequestExpertStatus.SENT)
                    .contains(re.getStatus()))
        .forEach(re -> re.updateStatus(RequestExpertStatus.FULFILLED));
    request.setStatus(RequestStatus.ACCEPTED);
    eventPublisher.publishEvent(new RequestAcceptedEvent(request, requestExpert.getExpert()));
  }

  @Transactional
  public void refuseRequest(
      Integer requestId, Integer expertId, RequestRefusalForm requestRefusalForm) {
    RequestExpert requestExpert =
        getRequestExpertIfInvolved(
            requestId, expertId, List.of(RequestExpertStatus.SENDING, RequestExpertStatus.SENT));
    Request request = requestExpert.getRequest();
    UserExpert userExpert = userExpertRepository.findByExpertId(expertId).orElseThrow();
    if (RefusalReason.UNAVAILABLE.equals(requestRefusalForm.getReason())) {
      var now = LocalDate.now(ZoneId.of("Europe/Paris"));
      userExpert.setUnavailableFrom(now);
      var until = requestRefusalForm.getUnavailableUntil();
      var unavailableUntil = Objects.nonNull(until) && !until.isAfter(now) ? null : until;
      userExpert.setUnavailableUntil(unavailableUntil);
      if (Objects.nonNull(unavailableUntil)) {
        requestRefusalForm.setJustification("jusqu'au " + unavailableUntil.format(dtfUnavailable));
      } else {
        requestRefusalForm.setJustification("pour une durée indéterminée");
      }
    }
    if (RefusalReason.GEO_CONSTRAINT.equals(requestRefusalForm.getReason())
        && StringUtils.isBlank(userExpert.getGeoConstraint())) {
      throw new AppException(
          "can't use GEO_CONSTRAINT reason if expert has no geo constraint defined");
    }
    if (RefusalReason.OTHER.equals(requestRefusalForm.getReason())
        && StringUtils.isBlank(requestRefusalForm.getJustification())) {
      throw new AppException("can't use OTHER reason if no justification");
    }
    requestExpert.updateStatusRefusal(
        requestRefusalForm.getReason(), requestRefusalForm.getJustification());
    sendRequestToNextExpert(request);
    checkAllExpertsRefused(request);
    eventPublisher.publishEvent(
        new RequestRefusedEvent(
            request,
            requestExpert.getExpert(),
            requestRefusalForm.getReason(),
            requestRefusalForm.getJustification()));
  }

  @Transactional
  @Scheduled(cron = "0 0 0 * * *", zone = "Europe/Paris")
  public void refuseRequestAuto() {
    log.info("refuseRequestAuto()");
    var now = LocalDateTime.now(ZoneId.of("Europe/Paris"));
    List<RequestExpert> awaitingAnswer =
        requestExpertRepository.findByStatusIn(RequestExpertStatus.SENT);
    Map<Request, List<RequestExpert>> requestExpertsPerRequest =
        awaitingAnswer.stream()
            .filter(re -> DateUtils.getDeadline(re.getSentOn(), weekdaysToReply).isBefore(now))
            .collect(groupingBy(RequestExpert::getRequest));
    requestExpertsPerRequest.forEach(
        (request, requestExpertList) -> {
          requestExpertList.forEach(
              re -> {
                re.updateStatus(RequestExpertStatus.REFUSED_AUTO);
                Optional.ofNullable(re.getExpert().getUserExpert())
                    .ifPresent(
                        userExpert -> {
                          userExpert.setUnavailableFrom(now.toLocalDate());
                          userExpert.setUnavailableUntil(null);
                        });
              });
          sendRequestToNextExpert(request);
          checkAllExpertsRefused(request);
          eventPublisher.publishEvent(
              new RequestRefusedAutoEvent(
                  request,
                  requestExpertList.stream()
                      .map(RequestExpert::getExpert)
                      .collect(Collectors.toList())));
        });
  }

  @Transactional
  public Note addQuestion(AddNote addNote, UserExpert user) {
    RequestExpert requestExpert =
        getRequestExpertIfInvolved(
            addNote.getRequestId(),
            user.getExpert().getId(),
            List.of(
                RequestExpertStatus.SENDING,
                RequestExpertStatus.SENT,
                RequestExpertStatus.ACCEPTED));
    Note question =
        Note.builder()
            .content(addNote.getContent())
            .request(requestExpert.getRequest())
            .createdOn(LocalDateTime.now())
            .user(user)
            .type(NoteType.QUESTION)
            .build();
    noteRepository.save(question);
    eventPublisher.publishEvent(
        new QuestionAskedEvent(requestExpert.getRequest(), user.getExpert()));
    return question;
  }

  @Transactional
  public void updateAvailability(Integer userExpertId, ExpertAvailabilityForm expertAvailability) {
    UserExpert userExpert = userExpertRepository.findById(userExpertId).orElseThrow();
    if (expertAvailability.getAvailable()) {
      userExpert.setUnavailableFrom(null);
      userExpert.setUnavailableUntil(null);
    } else {
      if (Objects.isNull(expertAvailability.getUnavailableFrom())) {
        expertAvailability.setUnavailableFrom(LocalDate.now(ZoneId.of("Europe/Paris")));
      }
      userExpert.setUnavailableFrom(expertAvailability.getUnavailableFrom());
      userExpert.setUnavailableUntil(expertAvailability.getUnavailableUntil());
    }
  }

  private RequestExpert getRequestExpertIfInvolved(
      Integer requestId, Integer expertId, List<RequestExpertStatus> statuses) {
    Optional<RequestExpert> optionalRequestExpert =
        requestExpertRepository.findById(new RequestExpertId(requestId, expertId));
    if (optionalRequestExpert.isPresent()) {
      RequestExpert requestExpert = optionalRequestExpert.get();
      if (CollectionUtils.isNotEmpty(statuses)) {
        if (statuses.contains(requestExpert.getStatus())
            && RequestStatus.OPEN.equals(requestExpert.getRequest().getStatus())
            && !requestExpert.getRequest().isArchived()) {
          return requestExpert;
        } else {
          throw new AppException("can't act on this request");
        }
      } else {
        return requestExpert;
      }
    }
    throw new AppException("expert not involved in this request");
  }

  private void sendRequestToNextExpert(Request request) {
    if (!request.isUrgent()) {
      request.getRequestExperts().stream()
          .filter(re -> RequestExpertStatus.QUEUED.equals(re.getStatus()))
          .findFirst()
          .ifPresent(re -> re.updateStatus(RequestExpertStatus.SENDING));
    }
  }

  private void checkAllExpertsRefused(Request request) {
    if (request.getRequestExperts().stream()
        .noneMatch(
            re ->
                List.of(RequestExpertStatus.SENDING, RequestExpertStatus.SENT)
                    .contains(re.getStatus()))) {
      request.setStatus(RequestStatus.REFUSED);
    }
  }

  @Transactional
  public String updateExpertData(Integer expertId, UpdatableData field, String value) {
    Expert expert = expertRepository.findById(expertId).orElseThrow();
    String cleanValue = TextUtils.cleanValue(value);
    switch (field) {
      case phoneMobile -> expert.setPhoneMobile(cleanValue);
      case phoneLandline -> expert.setPhoneLandline(cleanValue);
      case specialty -> expert.setSpecialty(cleanValue);
      case opalexe -> {
        if (Objects.isNull(expert.getUserExpert())) {
          throw new AppException("can't update opalexe");
        }
        expert.getUserExpert().setOpalexe(Boolean.valueOf(cleanValue));
      }
      case address -> expert.setAddress(cleanValue);
      case city -> expert.setCity(cleanValue);
      case zipCode -> expert.setZipCode(cleanValue);
      case geoConstraint -> {
        if (Objects.isNull(expert.getUserExpert())) {
          throw new AppException("can't update geoConstraint");
        }
        expert.getUserExpert().setGeoConstraint(cleanValue);
      }
      case email -> {
        if (Objects.nonNull(expert.getUserExpert())) {
          expert.getUserExpert().setEmail(cleanValue);
        }
        expert.setEmail(cleanValue);
      }
      case name -> {
        if (Objects.nonNull(expert.getUserExpert())) {
          expert.getUserExpert().setName(cleanValue);
        }
        expert.setName(cleanValue);
      }
      default -> throw new AppException("can't update field " + field);
    }
    return cleanValue;
  }

  @Transactional
  @Scheduled(cron = "0 0 0 * * *", zone = "Europe/Paris")
  public void checkAvailability() {
    log.info("checkAvailability()");
    // TODO write SQL query that returns only the relevant experts
    userExpertRepository.findByUnavailableFromIsNotNullOrUnavailableUntilIsNotNull().stream()
        .filter(e -> !Boolean.TRUE.equals(e.getExpert().getDeleted()))
        .forEach(
            e -> {
              if (Objects.nonNull(e.getUnavailableFrom())) {
                LocalDate now = LocalDate.now(ZoneId.of("Europe/Paris"));
                long daysSince = ChronoUnit.DAYS.between(e.getUnavailableFrom(), now);
                if (Objects.nonNull(e.getUnavailableUntil())
                    && e.getUnavailableUntil().isBefore(now)) {
                  log.info("userExpert {} is now available", e.getId());
                  eventPublisher.publishEvent(new ExpertIsNowAvailableEvent(e.getExpert()));
                  e.setUnavailableFrom(null);
                  e.setUnavailableUntil(null);
                } else if (Objects.isNull(e.getUnavailableUntil())
                    && (daysSince == 7
                        || daysSince == 14
                        || (daysSince % 30 == 0 && daysSince > 0))) {
                  log.info("userExpert {} is still unavailable ({} days)", e.getId(), daysSince);
                  eventPublisher.publishEvent(new ExpertIsStillUnavailableEvent(e.getExpert()));
                }
              } else { // so unavailableUntil is not null - should not happen, just in case...
                log.info("userExpert {} cleanup orphan availableUntil", e.getId());
                e.setUnavailableUntil(null);
              }
            });
  }
}
