package fr.gouv.beta.selexpert.service;

import com.amazonaws.AmazonServiceException;
import com.amazonaws.auth.AWSCredentials;
import com.amazonaws.auth.AWSStaticCredentialsProvider;
import com.amazonaws.auth.BasicAWSCredentials;
import com.amazonaws.client.builder.AwsClientBuilder.EndpointConfiguration;
import com.amazonaws.services.s3.AmazonS3;
import com.amazonaws.services.s3.AmazonS3ClientBuilder;
import com.amazonaws.services.s3.model.CopyObjectRequest;
import com.amazonaws.services.s3.model.DeleteObjectRequest;
import com.amazonaws.services.s3.model.ObjectMetadata;
import com.amazonaws.services.s3.model.S3Object;
import com.amazonaws.services.s3.model.S3ObjectInputStream;
import com.opencsv.CSVParserBuilder;
import com.opencsv.CSVReader;
import com.opencsv.CSVReaderBuilder;
import fr.gouv.beta.selexpert.AppException;
import fr.gouv.beta.selexpert.model.ExpertRawData;
import jakarta.annotation.PostConstruct;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.Arrays;
import java.util.List;
import java.util.Objects;
import java.util.stream.Collectors;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Service;
import org.springframework.web.multipart.MultipartFile;
import org.springframework.web.servlet.mvc.method.annotation.StreamingResponseBody;

@Service
@Slf4j
public class FileService {

  private AmazonS3 s3client;

  @Value("${selexpert.s3.endpoint}")
  private String endpoint;

  @Value("${selexpert.s3.bucket}")
  private String bucket;

  @Value("${selexpert.s3.accessKey}")
  private String accessKey;

  @Value("${selexpert.s3.secretKey}")
  private String secretKey;

  @PostConstruct
  private void initS3() {
    AWSCredentials credentials = new BasicAWSCredentials(this.accessKey, this.secretKey);
    this.s3client =
        AmazonS3ClientBuilder.standard()
            .withCredentials(new AWSStaticCredentialsProvider(credentials))
            .withEndpointConfiguration(new EndpointConfiguration(endpoint, null))
            .build();
  }

  public void upload(MultipartFile file, String fileName) {
    ObjectMetadata metadata = new ObjectMetadata();
    metadata.setContentType(file.getContentType());
    metadata.setContentLength(file.getSize());
    try {
      s3client.putObject(bucket, fileName, file.getInputStream(), metadata);
    } catch (IOException e) {
      log.error("s3 upload", e);
      throw new AppException("can't upload file to s3: " + file.getOriginalFilename());
    }
  }

  public StreamingResponseBody download(String fileName) {
    S3Object s3object = s3client.getObject(bucket, fileName);
    S3ObjectInputStream inputStream = s3object.getObjectContent();
    return outputStream -> {
      int numberOfBytesToWrite;
      byte[] data = new byte[1024];
      while ((numberOfBytesToWrite = inputStream.read(data, 0, data.length)) != -1) {
        outputStream.write(data, 0, numberOfBytesToWrite);
      }
      inputStream.close();
    };
  }

  public void delete(String fileName) {
    try {
      log.info("delete {}", fileName);
      s3client.deleteObject(new DeleteObjectRequest(bucket, fileName));
    } catch (AmazonServiceException e) {
      log.error("s3 delete file " + fileName, e);
    }
  }

  public boolean copy(String fileNameFrom, String fileNameTo) {
    log.info("copy {} -> {}", fileNameFrom, fileNameTo);
    try {
      s3client.copyObject(new CopyObjectRequest(bucket, fileNameFrom, bucket, fileNameTo));
      return true;
    } catch (AmazonServiceException e) {
      log.error("s3 copy file", e);
    }
    return false;
  }

  public List<ExpertRawData> getExpertRawDataFromFile(InputStream fileInputStream) {
    try (CSVReader csvReader =
        new CSVReaderBuilder(new BufferedReader(new InputStreamReader(fileInputStream)))
            .withCSVParser(
                new CSVParserBuilder().withSeparator(';').withIgnoreQuotations(true).build())
            .build()) {
      String[] headerRow = csvReader.readNext();
      if (Objects.isNull(headerRow)) {
        throw new AppException("EMPTY FILE");
      }
      if (!ExpertRawData.FileHeader.headers().equals(Arrays.asList(headerRow))) {
        throw new AppException("ROWS INVALID");
      }
      return csvReader.readAll().stream().map(ExpertRawData::fromRow).collect(Collectors.toList());
    } catch (Exception e) {
      throw new AppException("CAN'T READ FILE");
    }
  }
}
