package fr.gouv.beta.selexpert.service;

import static java.util.stream.Collectors.groupingBy;

import fr.gouv.beta.selexpert.model.EnversRevision;
import fr.gouv.beta.selexpert.model.Expert;
import fr.gouv.beta.selexpert.model.ExpertExpertiseNew;
import fr.gouv.beta.selexpert.model.ExpertNew;
import fr.gouv.beta.selexpert.model.ExpertNewExpertiseNew;
import fr.gouv.beta.selexpert.model.ExpertRawData;
import fr.gouv.beta.selexpert.model.ExpertType;
import fr.gouv.beta.selexpert.model.ExpertiseNew;
import fr.gouv.beta.selexpert.model.UserAdmin;
import fr.gouv.beta.selexpert.repository.AppealCourtRepository;
import fr.gouv.beta.selexpert.repository.AuditRepository;
import fr.gouv.beta.selexpert.repository.ExpertNewRepository;
import fr.gouv.beta.selexpert.repository.ExpertRawDataRepository;
import fr.gouv.beta.selexpert.repository.ExpertRepository;
import fr.gouv.beta.selexpert.repository.ExpertiseNewRepository;
import fr.gouv.beta.selexpert.repository.ExpertiseRepository;
import fr.gouv.beta.selexpert.util.TextUtils;
import jakarta.transaction.Transactional;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;
import java.util.Scanner;
import java.util.Set;
import java.util.function.Consumer;
import java.util.function.Function;
import java.util.stream.Collectors;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.collections4.IterableUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.math.NumberUtils;
import org.hibernate.envers.RevisionType;
import org.slf4j.helpers.MessageFormatter;
import org.springframework.stereotype.Service;

@Service
@Slf4j
@RequiredArgsConstructor
public class ImportService {

  private final ExpertRawDataRepository expertRawDataRepository;
  private final ExpertNewRepository expertNewRepository;
  private final ExpertiseNewRepository expertiseNewRepository;
  private final ExpertiseRepository expertiseRepository;
  private final AppealCourtRepository appealCourtRepository;
  private final AuditRepository auditRepository;
  private final ExpertRepository expertRepository;
  private final FileService fileService;

  /*
  1. import csv file data into expert raw data
  2. insert/update(all date)/delete(flag) expertNew table
  3. (TEMP) try to link expert to expertNew
  4. insert/update(categories only)/delete(flag) expert based on expertNew


  TODO: manually added and honorary => update to new categories
   */

  @Transactional
  public String importExpertRawData(InputStream fileInputStream, boolean dryRun, UserAdmin admin) {
    /*
    import data from the file as is
    detect changes to create a new revision (audit envers)
    set the flag if there are any changes
     */
    if (admin.isImporting()) return "erreur : import en cours";
    StringBuilder report = new StringBuilder();

    List<ExpertRawData> expertRawDataFromFile =
        fileService.getExpertRawDataFromFile(fileInputStream);
    String idCa = expertRawDataFromFile.get(0).getIdCa();
    String ca = expertRawDataFromFile.get(0).getCa();
    List<ExpertRawData> expertRawDataFromDB =
        expertRawDataRepository.getByIdCaOrderByIdExpertAscIdExpertCategorieAsc(idCa);
    if (dryRun) addToReport(report, "DRY RUN\n");
    addToReport(report, "CA : {} ({})", ca, idCa);
    addToReport(report, "{} lines in the file", expertRawDataFromFile.size());
    addToReport(report, "{} lines in the DB", expertRawDataFromDB.size());

    // remove the unchanged
    List<ExpertRawData> expertRawDataNoChange =
        expertRawDataFromFile.stream().filter(expertRawDataFromDB::contains).toList();
    expertRawDataFromFile.removeAll(expertRawDataNoChange);
    expertRawDataFromDB.removeAll(expertRawDataNoChange);
    addToReport(report, "{} lines unchanged", expertRawDataNoChange.size());

    // insert / update
    List<List<ExpertRawData>> insertOrUpdateExpertRawData =
        IterableUtils.partition(
            expertRawDataFromFile,
            expertRawData -> expertRawData.isContainedIn(expertRawDataFromDB));
    List<ExpertRawData> insertExpertRawData = insertOrUpdateExpertRawData.get(1);
    List<ExpertRawData> updateExpertRawData = insertOrUpdateExpertRawData.get(0);
    addToReport(report, "{} to insert", insertExpertRawData.size());
    if (dryRun) insertExpertRawData.forEach(d -> addToReport(report, "\n{}\n", d));
    addToReport(report, "{} to update", updateExpertRawData.size());
    if (dryRun) updateExpertRawData.forEach(d -> addToReport(report, "\n{}\n", d));
    if (!dryRun && (!insertExpertRawData.isEmpty() || !updateExpertRawData.isEmpty())) {
      expertRawDataRepository.saveAll(expertRawDataFromFile);
      admin.setImporting(true);
      addToReport(report, "==> {} INSERTED / UPDATED", expertRawDataFromFile.size());
    }

    // delete
    expertRawDataFromDB.removeIf(expertRawData -> expertRawData.isContainedIn(updateExpertRawData));
    addToReport(report, "{} to delete", expertRawDataFromDB.size());
    if (dryRun) expertRawDataFromDB.forEach(d -> addToReport(report, "\n{}\n", d));
    if (!dryRun && !expertRawDataFromDB.isEmpty()) {
      expertRawDataRepository.deleteAll(expertRawDataFromDB);
      admin.setImporting(true);
      addToReport(report, "==> {} DELETED", expertRawDataFromDB.size());
    }

    return report.toString();
  }

  @Transactional
  public String buildExperts(boolean dryRun, UserAdmin admin) {
    /*
    can be executed only if there are new changes
    get all the idExpert's related to the changes
    compare expert as built from the new file data VS the one in the DB
    insert and delete are straightforward
    if update, updating the list of expertises is straightforward
    for the expert data, make sure that the data from the file does not override the data from
    the DB if it was manually modified by the expert or the admin
     */
    if (!admin.isImporting()) return "erreur : pas d'import en cours";
    StringBuilder report = new StringBuilder();
    if (dryRun) addToReport(report, "DRY RUN\n");

    Map<String, ExpertiseNew> expertiseMap =
        expertiseNewRepository.findAll().stream()
            .collect(Collectors.toMap(ExpertiseNew::getCode, Function.identity()));
    List<String> changedIdExperts = changedIdExperts();
    addToReport(
        report, "there are changes for experts with idExpert (esaboraId): {}", changedIdExperts);
    changedIdExperts.forEach(
        idExpert -> {
          Optional<ExpertNew> expertFromDB = expertNewRepository.findByEsaboraId(idExpert);
          Optional<ExpertNew> expertFromRawData =
              buildExpert(expertRawDataRepository.getByIdExpert(idExpert), expertiseMap, report);

          // update
          if (expertFromDB.isPresent() && expertFromRawData.isPresent()) {
            addToReport(report, "\nupdate expertNew with esaboraId {}", idExpert);
            if (dryRun)
              addToReport(
                  report,
                  "from RAW DATA :\n{}\nfrom DB :\n{}",
                  expertFromRawData.get(),
                  expertFromDB.get());
            updateExpert(expertFromDB.get(), expertFromRawData.get(), report, dryRun);
          } else { // either insert or delete
            expertFromRawData.ifPresent(
                expert -> {
                  addToReport(report, "insert expertNew with esaboraId {}", idExpert);
                  if (dryRun) addToReport(report, "{}", expert);
                  if (!dryRun) expertNewRepository.save(expert);
                });
            expertFromDB.ifPresent(
                expert -> {
                  addToReport(report, "delete (flag) expertNew with esaboraId {}", idExpert);
                  if (dryRun) addToReport(report, "{}", expert);
                  if (!dryRun) {
                    expert.setDeleted(true);
                  }
                });
          }
        });
    if (!dryRun) admin.setImporting(false);
    return report.toString();
  }

  public void expertDiff() {
    // TODO
  }

  @Transactional
  public String migration(Integer courtId, boolean auto) {
    /*
    try to match existing experts to expertNew (those who don't have a link yet)
    match auto if found through email or full name
    match manually (prompt) if found otherwise
     */
    StringBuilder report = new StringBuilder();
    List<ExpertNew> expertNewList = expertNewRepository.findByAppealCourtId(courtId);
    List<Expert> expertList =
        expertRepository.findByAppealCourtIdAndExpertNewIsNullOrderById(courtId);
    addToReport(report, "courtId : {}", courtId);
    addToReport(report, "AUTO ? {}", auto);
    addToReport(report, "still {} experts to match...", expertList.size());
    int count = 0;
    for (Expert expert : expertList) {
      addToReport(report, "\n\n\n\n\n\n*** {} {}", expert.getName(), expert.getEmail());
      System.out.println(++count + "/" + expertList.size());
      if (Boolean.TRUE.equals(expert.getHonorary())) {
        addToReport(report, auto, "SKIP: HONORARY");
        continue;
      }

      // AUTO
      // if match by email or full name, save automatically
      ExpertNew found =
          findNewExpert(
              "EMAIL",
              expert.getEmail(),
              expertNewList.stream().collect(groupingBy(ExpertNew::getEmail)),
              report,
              auto);
      if (auto && Objects.nonNull(found)) {
        expert.setExpertNew(found);
        continue;
      }
      found =
          findNewExpert(
              "FULL NAME",
              expert.getName(),
              expertNewList.stream()
                  .collect(groupingBy(en -> en.getName() + " " + en.getFirstName())),
              report,
              auto);
      if (auto && Objects.nonNull(found)) {
        expert.setExpertNew(found);
        continue;
      }
      if (auto) addToReport(report, "FAIL");

      // MANUAL
      if (!auto) {

        Scanner scanner = new Scanner(System.in);

        if (confirmedSave(
            findNewExpert(
                "LAST NAME",
                extractLastName(expert.getName().split(" ")),
                expertNewList.stream().collect(groupingBy(ExpertNew::getName)),
                report,
                false),
            expert,
            scanner,
            report)) continue;

        if (expert.getName().contains("épouse")) {
          if (confirmedSave(
              findNewExpert(
                  "EPOUSE",
                  extractLastName((expert.getName().split("épouse ")[1]).split(" ")),
                  expertNewList.stream().collect(groupingBy(ExpertNew::getName)),
                  report,
                  false),
              expert,
              scanner,
              report)) continue;
        }
        if (expert.getName().contains("née")) {
          if (confirmedSave(
              findNewExpert(
                  "NEE",
                  extractLastName((expert.getName().split("née ")[1]).split(" ")),
                  expertNewList.stream().collect(groupingBy(ExpertNew::getName)),
                  report,
                  false),
              expert,
              scanner,
              report)) continue;
        }

        System.out.println("ENTER expert_new.esaboraId");
        System.out.println("[ 0 delete | 1 skip | 9 finish ]");
        if (Boolean.TRUE.equals(expert.getDeleted()))
          addToReport(report, true, "WARNING: expert is DELETED (but try to match anyway)");
        if (Boolean.TRUE.equals(expert.getManuallyAdded()))
          addToReport(report, true, "WARNING: expert is MANUALLY ADDED (skip if not found)");
        String esaboraId = scanner.next();
        ExpertNew manuallyFound = null;
        while (!List.of("0", "1", "9").contains(esaboraId) && Objects.isNull(manuallyFound)) {
          Optional<ExpertNew> maybeFound = expertNewRepository.findByEsaboraId(esaboraId);
          if (maybeFound.isPresent()) manuallyFound = maybeFound.get();
          else {
            System.out.println("not found, try again");
            esaboraId = scanner.next();
          }
        }
        if (Objects.nonNull(manuallyFound)) {
          addToReport(report, true, "--- SAVE {}", esaboraId);
          addToReport(
              report,
              true,
              "{} {} {}",
              manuallyFound.getName(),
              manuallyFound.getFirstName(),
              manuallyFound.getEmail());
        } else if (esaboraId.equals("0")) {
          addToReport(report, true, "--- DELETE (flag)");
          expert.setDeleted(true);
        } else if (esaboraId.equals("1")) {
          addToReport(report, true, "--- SKIP");
        } else {
          addToReport(report, true, "--- EXIT AND COMMIT CHANGES");
          break;
        }
      }
    }
    return report.toString();
  }

  @Transactional
  public String updateExperts(Integer courtId) {
    /*
    1. for experts who are linked to an expertNew, update categories and deleted flag
    2. add new experts (expertNew that are not linked to any expert)
     */
    StringBuilder report = new StringBuilder();
    List<Expert> expertsLinked = expertRepository.findByAppealCourtIdAndExpertNewIsNotNull(courtId);

    // 1
    expertsLinked.forEach(
        expertLinked -> {
          if (!Boolean.valueOf(expertLinked.getExpertNew().isDeleted())
              .equals(expertLinked.getDeleted())) {
            addToReport(report, "update deleted flag for {}", expertLinked.getId());
            expertLinked.setDeleted(expertLinked.getExpertNew().isDeleted());
          }
          if (!expertLinked.getExpertNew().isDeleted())
            updateExpertises(expertLinked, expertLinked.getExpertNew(), report);
        });

    // 2
    List<String> emails =
        expertRepository.findByAppealCourtIdAndEmailIsNotNull(courtId).stream()
            .map(Expert::getEmail)
            .filter(StringUtils::isNotBlank)
            .toList();
    expertNewRepository
        .findByAppealCourtIdAndIdNotIn(
            courtId, expertsLinked.stream().map(e -> e.getExpertNew().getId()).toList())
        .forEach(
            expertNew -> {
              if (emails.contains(expertNew.getEmail())) {
                addToReport(report, "ERREUR email exists: {} DO NOT SAVE " + expertNew.getEmail());
                return;
              }
              Expert expert = buildExpertFromExpertNew(expertNew);
              expert.setExpertNew(expertNew);
              expertRepository.save(expert);
            });
    return report.toString();
  }

  private Expert buildExpertFromExpertNew(ExpertNew expertNew) {
    Expert expert =
        Expert.builder()
            .appealCourt(expertNew.getAppealCourt())
            .title(
                expertNew.getType() != ExpertType.COMPANY ? expertNew.getType().getLabel() : null)
            .name(
                expertNew.getName()
                    + (StringUtils.isNotBlank(expertNew.getFirstName())
                        ? (" " + expertNew.getFirstName())
                        : ""))
            .email(StringUtils.isNotBlank(expertNew.getEmail()) ? expertNew.getEmail() : null)
            .phoneMobile(expertNew.getPhoneMobile())
            .phoneLandline(expertNew.getPhoneLandline())
            .address(expertNew.getAddress())
            .zipCode(expertNew.getZipCode())
            .city(expertNew.getCity())
            .build();
    expert.setExpertExpertisesNew(
        expertNew.getExpertExpertises().stream()
            .map(
                expertNewExpertiseNew ->
                    new ExpertExpertiseNew(
                        expert,
                        expertNewExpertiseNew.getExpertise(),
                        expertNewExpertiseNew.getSpecialty()))
            .collect(Collectors.toSet()));
    return expert;
  }

  private boolean confirmedSave(
      ExpertNew found, Expert expert, Scanner scanner, StringBuilder report) {
    if (Objects.nonNull(found)) {
      addToReport(report, true, "MAYBE FOUND ???");
      System.out.println("type y to confirm or n to cancel");
      String answer = scanner.next();
      while (!List.of("y", "n").contains(answer)) answer = scanner.next();
      if (answer.equals("y")) {
        expert.setExpertNew(found);
        addToReport(report, true, "--- confirmed and saved");
        return true;
      } else {
        addToReport(report, true, "--- cancelled");
      }
    }
    return false;
  }

  private ExpertNew findNewExpert(
      String by,
      String value,
      Map<String, List<ExpertNew>> map,
      StringBuilder report,
      boolean auto) {
    if (StringUtils.isNotBlank(value) && map.containsKey(value)) {
      if (map.get(value).size() == 1) {
        addToReport(report, !auto, "OK: FOUND BY {}", by);
        ExpertNew expertNew = map.get(value).get(0);
        addToReport(
            report,
            !auto,
            "{} {} {} {}",
            expertNew.getName(),
            expertNew.getFirstName(),
            expertNew.getEmail(),
            expertNew.getEsaboraId());
        return expertNew;
      } else {
        addToReport(report, !auto, "KO: FOUND MULTIPLE {}", by);
        map.get(value)
            .forEach(
                expertNew ->
                    addToReport(
                        report,
                        !auto,
                        expertNew.getName()
                            + " "
                            + expertNew.getFirstName()
                            + " "
                            + expertNew.getEmail()));
        return null;
      }
    }
    addToReport(report, !auto, "KO: NOT FOUND BY {}", by);
    return null;
  }

  private String extractLastName(String[] nameChunks) {
    String lastName = "";
    for (String chunk : nameChunks) {
      if (chunk.matches("[A-ZÀÈÉÊËÎÏ\\-]+")) lastName += chunk + " ";
      else break;
    }
    return lastName.trim();
  }

  private Optional<ExpertNew> buildExpert(
      List<ExpertRawData> expertRawDataForOneExpert,
      Map<String, ExpertiseNew> expertiseMap,
      StringBuilder report) {
    if (CollectionUtils.isEmpty(expertRawDataForOneExpert)) return Optional.empty();
    ExpertRawData e = expertRawDataForOneExpert.get(0);
    ExpertNew expert =
        ExpertNew.builder()
            .appealCourt(appealCourtRepository.getAppealCourtByCode(e.getIdCa()).orElseThrow())
            .esaboraId(e.getIdExpert())
            .type(ExpertType.fromLabel(e.getCivilite()))
            .name(StringUtils.isNotBlank(e.getNomUsage()) ? e.getNomUsage() : e.getNom())
            .firstName(e.getPrenom())
            .address(
                TextUtils.join(
                    "\n", e.getAdresse1(), e.getAdresse2(), e.getAdresse3(), e.getLieuDit()))
            .zipCode(e.getCodePostal())
            .city(e.getVille())
            .email(e.getCourriel())
            .phoneMobile(e.getMobile().replaceAll("\\D+", ""))
            .phoneLandline(e.getFixe().replaceAll("\\D+", ""))
            .phoneFax(e.getFax().replaceAll("\\D+", ""))
            .build();
    expert.setExpertExpertises(
        expertRawDataForOneExpert.stream()
            .map(
                expertRawData -> {
                  String code =
                      expertRawData.getCategorie().replace("-", ".").replaceAll("\\.0", ".") + ".";
                  ExpertiseNew expertise = expertiseMap.get(code);
                  if (Objects.nonNull(expertise)) {
                    return new ExpertNewExpertiseNew(
                        expert,
                        expertise,
                        TextUtils.join(
                            "\n",
                            expertRawData.getSpecialite(),
                            expertRawData.getSousSpecialite()));
                  }
                  addToReport(
                      report,
                      "ERREUR expertise non trouvée pour {} : {} {}",
                      expert.getName() + " " + expert.getFirstName(),
                      code,
                      expertRawData.getRubrique());
                  addToReport(
                      report,
                      "ancienne expertise ? {}",
                      expertiseRepository.findByCode(expertRawData.getCategorie()).isPresent());
                  return null;
                })
            .filter(Objects::nonNull)
            .collect(Collectors.toSet()));
    return Optional.of(expert);
  }

  private List<String> changedIdExperts() {
    // get all the ExpertRawData that changed after last import
    List<Integer> revisionNumbers = auditRepository.getRevisionNumbers(ExpertRawData.class);
    if (CollectionUtils.isEmpty(revisionNumbers)) return List.of();
    return auditRepository.getRevision(revisionNumbers.get(0), ExpertRawData.class).stream()
        .map(
            expertRawData ->
                Optional.ofNullable(expertRawData.getIdExpert())
                    .orElseGet(
                        () -> {
                          // no idExpert means deletion, get the idExpert from previous version
                          return auditRepository
                              .getRevisionFor(
                                  expertRawData.getIdExpertCategorie(),
                                  revisionNumbers.get(1),
                                  ExpertRawData.class)
                              .getIdExpert();
                        }))
        .distinct()
        .toList();
  }

  private void updateExpert(
      ExpertNew expertFromDB, ExpertNew expertFromRawData, StringBuilder report, boolean dryRun) {

    updateField(
        expertFromDB.getName(),
        expertFromRawData.getName(),
        expertFromDB::setName,
        "name",
        expertFromDB.getId(),
        report,
        dryRun);
    updateField(
        expertFromDB.getFirstName(),
        expertFromRawData.getFirstName(),
        expertFromDB::setFirstName,
        "firstName",
        expertFromDB.getId(),
        report,
        dryRun);
    updateField(
        expertFromDB.getAddress(),
        expertFromRawData.getAddress(),
        expertFromDB::setAddress,
        "address",
        expertFromDB.getId(),
        report,
        dryRun);
    updateField(
        expertFromDB.getEmail(),
        expertFromRawData.getEmail(),
        expertFromDB::setEmail,
        "email",
        expertFromDB.getId(),
        report,
        dryRun);
    updateField(
        expertFromDB.getPhoneMobile(),
        expertFromRawData.getPhoneMobile(),
        expertFromDB::setPhoneMobile,
        "phoneMobile",
        expertFromDB.getId(),
        report,
        dryRun);
    updateField(
        expertFromDB.getPhoneLandline(),
        expertFromRawData.getPhoneLandline(),
        expertFromDB::setPhoneLandline,
        "phoneLandline",
        expertFromDB.getId(),
        report,
        dryRun);
    updateField(
        expertFromDB.getPhoneFax(),
        expertFromRawData.getPhoneFax(),
        expertFromDB::setPhoneFax,
        "phoneFax",
        expertFromDB.getId(),
        report,
        dryRun);

    updateExpertises(expertFromDB, expertFromRawData, report, dryRun);
  }

  private void updateField(
      String valueFromDB,
      String valueFromRawData,
      Consumer<String> fieldSetter,
      String fieldName,
      Integer checkOverrideExpertId,
      StringBuilder report,
      boolean dryRun) {
    if (!valueFromDB.equals(valueFromRawData)) {
      String lastModifiedBy = lastModifiedBy(fieldName, checkOverrideExpertId);
      if (Objects.nonNull(lastModifiedBy) && !lastModifiedBy.equals("import")) {
        addToReport(
            report,
            "\n**DO NOT UPDATE** {}:\n{}\n=>\n{}\n (modified by {})",
            fieldName,
            valueFromDB,
            valueFromRawData,
            lastModifiedBy);
      } else {
        addToReport(report, "\nUPDATE {}:\n{}\n=>\n{}", fieldName, valueFromDB, valueFromRawData);
        if (!dryRun) fieldSetter.accept(valueFromRawData);
      }
    }
  }

  private String lastModifiedBy(String field, Integer expertId) {
    String lastModifiedBy = null;
    var revisions = auditRepository.getRevisionsFor(expertId, ExpertNew.class);
    Collections.reverse(revisions);
    for (Object revision : revisions) {
      Object[] revisionArray = (Object[]) revision;
      //      final ExpertNew entity = (ExpertNew) revisionArray[0];
      final EnversRevision enversRevision = (EnversRevision) revisionArray[1];
      final RevisionType revisionType = (RevisionType) revisionArray[2];
      final Set<String> propertiesChanged = (Set<String>) revisionArray[3];
      if (revisionType != RevisionType.MOD || !propertiesChanged.contains(field)) continue;
      lastModifiedBy = enversRevision.getUsername();
      break;
    }
    return lastModifiedBy;
  }

  private void updateExpertises(
      ExpertNew expertFromDB, ExpertNew expertFromRawData, StringBuilder report, boolean dryRun) {
    var expertFromDBExpertiseStrList = expertExpertiseToStrList(expertFromDB);
    var expertFromRawDataExpertiseStrList = expertExpertiseToStrList(expertFromRawData);
    if (expertFromDBExpertiseStrList.equals(expertFromRawDataExpertiseStrList)) return;
    addToReport(
        report,
        "\nEXPERTISES CHANGE :\n{}\n{}",
        expertFromDBExpertiseStrList,
        expertFromRawDataExpertiseStrList);
    if (!dryRun) {
      var expertExpertisesFromDB = expertFromDB.getExpertExpertises();
      var expertExpertisesFromRawData =
          expertFromRawData.getExpertExpertises().stream()
              .map(
                  ee ->
                      new ExpertNewExpertiseNew(expertFromDB, ee.getExpertise(), ee.getSpecialty()))
              .collect(Collectors.toSet());
      Iterator<ExpertNewExpertiseNew> it = expertExpertisesFromDB.iterator();
      while (it.hasNext()) {
        var expertExpertiseFromDB = it.next();
        if (!expertExpertisesFromRawData.contains(expertExpertiseFromDB)) {
          it.remove();
        } else {
          for (ExpertNewExpertiseNew expertExpertiseFromRawData : expertExpertisesFromRawData) {
            if (expertExpertiseFromDB.equals(expertExpertiseFromRawData)) {
              expertExpertiseFromDB.setSpecialty(expertExpertiseFromRawData.getSpecialty());
              break;
            }
          }
        }
      }
      expertExpertisesFromDB.addAll(expertExpertisesFromRawData); // it's a set
    }
  }

  private void updateExpertises(Expert expert, ExpertNew expertNew, StringBuilder report) {
    // TODO duplicated code!!
    var expertExpertiseNewStrList = expertExpertiseToStrList(expert);
    var expertNewExpertiseNewStrList = expertExpertiseToStrList(expertNew);
    if (expertExpertiseNewStrList.equals(expertNewExpertiseNewStrList)) return;
    addToReport(
        report,
        "\nEXPERTISES CHANGE :\n{}\n{}",
        expertExpertiseNewStrList,
        expertNewExpertiseNewStrList);
    var expertExpertises = expert.getExpertExpertisesNew();

    Iterator<ExpertExpertiseNew> it = expertExpertises.iterator();
    while (it.hasNext()) {
      var expertExpertiseNew = it.next();
      // delete
      if (!expertNew.getExpertExpertises().stream()
          .map(expertNewExpertiseNew -> expertNewExpertiseNew.getExpertise().getId())
          .toList()
          .contains(expertExpertiseNew.getExpertise().getId())) {
        it.remove();
      } else {
        // update specialty
        for (ExpertNewExpertiseNew expertNewExpertise : expertNew.getExpertExpertises()) {
          if (expertExpertiseNew
              .getExpertise()
              .getId()
              .equals(expertNewExpertise.getExpertise().getId())) {
            expertExpertiseNew.setSpecialty(expertNewExpertise.getSpecialty());
            break;
          }
        }
      }
    }

    // add new
    var listOfIds =
        expertExpertises.stream()
            .map(expertExpertiseNew -> expertExpertiseNew.getExpertise().getId())
            .toList();
    expertExpertises.addAll(
        expertNew.getExpertExpertises().stream()
            .filter(
                expertNewExpertiseNew ->
                    !listOfIds.contains(expertNewExpertiseNew.getExpertise().getId()))
            .map(
                expertNewExpertiseNew ->
                    new ExpertExpertiseNew(
                        expert,
                        expertNewExpertiseNew.getExpertise(),
                        expertNewExpertiseNew.getSpecialty()))
            .toList());
  }

  private List<String> expertExpertiseToStrList(ExpertNew expert) {
    return expert.getExpertExpertises().stream()
        .map(
            expertExpertise ->
                expertExpertise.getExpertise().getCode() + " " + expertExpertise.getSpecialty())
        .sorted()
        .toList();
  }

  private List<String> expertExpertiseToStrList(Expert expert) {
    // TODO duplicated code!!
    return expert.getExpertExpertisesNew().stream()
        .map(
            expertExpertise ->
                expertExpertise.getExpertise().getCode() + " " + expertExpertise.getSpecialty())
        .sorted()
        .toList();
  }

  private void addToReport(StringBuilder report, boolean sout, String text, Object... data) {
    String message = MessageFormatter.arrayFormat(text, data).getMessage();
    if (sout) System.out.println(message);
    else log.info("*** " + message);
    report.append(message).append("\n");
  }

  private void addToReport(StringBuilder report, String text, Object... data) {
    addToReport(report, false, text, data);
  }

  @Transactional
  public String importTemp(InputStream fileInputStream, boolean dryRun) {
    StringBuilder report = new StringBuilder();
    List<ExpertRawData> expertRawDataFromFile =
        fileService.getExpertRawDataFromFile(fileInputStream);
    String idCa = expertRawDataFromFile.get(0).getIdCa();
    String ca = expertRawDataFromFile.get(0).getCa();
    if (dryRun) addToReport(report, "DRY RUN\n");
    addToReport(report, "CA : {} ({})", ca, idCa);
    addToReport(report, "{} lines in the file", expertRawDataFromFile.size());

    Map<String, ExpertiseNew> expertiseMap =
        expertiseNewRepository.findAll().stream()
            .collect(Collectors.toMap(ExpertiseNew::getCode, Function.identity()));

    String name = "";
    List<ExpertRawData> dataForOneExpert = new ArrayList<>();
    List<ExpertNew> expertNewList = new ArrayList<>();
    for (int i = 0; i <= expertRawDataFromFile.size(); i++) {
      var row = i < expertRawDataFromFile.size() ? expertRawDataFromFile.get(i) : null;
      if (Objects.isNull(row) || !row.getNom().equals(name)) {
        if (!"".equals(name)) {
          ExpertNew expertNew = buildExpert(dataForOneExpert, expertiseMap, report).orElseThrow();
          expertNew.setApplication2024(
              dataForOneExpert.stream().allMatch(d -> "01/01/2024".equals(d.getDateApplication())));
          expertNewList.add(expertNew);
        }
        if (Objects.isNull(row)) break;
        name = row.getNom();
        dataForOneExpert.clear();
      }
      dataForOneExpert.add(row);
    }

    if (true) {
      if (!dryRun)
        expertNewList.stream()
            .map(this::buildExpertFromExpertNew)
            .forEach(
                expert -> {
                  if (Objects.isNull(expert.getEmail())) {
                    expertRepository.save(expert);
                    addToReport(report, "EMAIL VIDE {}", expert.getName());
                  } else
                    expertRepository
                        .findByEmail(expert.getEmail())
                        .ifPresentOrElse(
                            existing ->
                                addToReport(
                                    report,
                                    "EMAIL DOUBLON {} {} (existant {} {}) ",
                                    expert.getEmail(),
                                    expert.getName(),
                                    existing.getEmail(),
                                    existing.getName()),
                            () -> {
                              addToReport(
                                  report,
                                  "NOUVEL EXPERT {} {}",
                                  expert.getEmail(),
                                  expert.getName());
                              expertRepository.save(expert);
                            });
                });
      return report.toString();
    }

    // add new appeals courts
    if (!List.of("100011", "100017", "100018").contains(idCa)) {
      if (!dryRun)
        expertRepository.saveAll(
            expertNewList.stream().map(this::buildExpertFromExpertNew).toList());
      return report.toString();
    }

    List<Expert> expertList =
        expertRepository.findByAppealCourtIdAndExpertNewIsNullOrderById(
            appealCourtRepository.getAppealCourtByCode(idCa).orElseThrow().getId());

    addToReport(report, "{} expertNewList", expertNewList.size());
    addToReport(report, "find by EMAIL");
    Iterator<ExpertNew> expertNewIterator = expertNewList.iterator();
    while (expertNewIterator.hasNext()) {
      ExpertNew expertNew = expertNewIterator.next();
      Expert expert =
          IterableUtils.find(
              expertList,
              e ->
                  StringUtils.isNotBlank(e.getEmail())
                      && e.getEmail().equals(expertNew.getEmail()));
      if (Objects.nonNull(expert)) {
        if (!dryRun) updateExpertises(expert, expertNew, report);
        expertNewIterator.remove();
      }
    }
    addToReport(report, "{} expertNewList", expertNewList.size());
    addToReport(report, "find by FULL NAME");
    expertNewIterator = expertNewList.iterator();
    while (expertNewIterator.hasNext()) {
      ExpertNew expertNew = expertNewIterator.next();
      var expertNewFullName =
          expertNew.getName()
              + (StringUtils.isNotBlank(expertNew.getFirstName())
                  ? " " + expertNew.getFirstName()
                  : "");
      var matchCount =
          IterableUtils.countMatches(expertList, e -> e.getName().equals(expertNewFullName));
      if (matchCount != 1) {
        if (matchCount > 1)
          addToReport(report, "error: multiple full name matches: {}", expertNewFullName);
        continue;
      }
      Expert expert = IterableUtils.find(expertList, e -> e.getName().equals(expertNewFullName));
      if (Objects.nonNull(expert)) {
        if (!dryRun) updateExpertises(expert, expertNew, report);
        expertNewIterator.remove();
      }
    }
    addToReport(report, "{} expertNewList", expertNewList.size());
    addToReport(report, "add application 01/01/2024");
    expertNewIterator = expertNewList.iterator();
    while (expertNewIterator.hasNext()) {
      ExpertNew expertNew = expertNewIterator.next();
      if (expertNew.isApplication2024()) {
        if (!dryRun) expertRepository.save(buildExpertFromExpertNew(expertNew));
        expertNewIterator.remove();
      }
    }
    addToReport(report, "{} expertNewList", expertNewList.size());
    addToReport(report, "MANUAL INPUT");
    Scanner scanner = new Scanner(System.in);
    for (ExpertNew expertNew : expertNewList) {
      addToReport(
          report,
          true,
          "\n\n{} {} {}",
          expertNew.getName(),
          expertNew.getFirstName(),
          expertNew.getEmail());
      if (!dryRun) {
        System.out.println("ENTER matching expert.Id");
        System.out.println("[ 0 to create new expert ]");
        Expert manuallyFound = null;
        String expertId = scanner.next();
        while (!expertId.equals("0") && Objects.isNull(manuallyFound)) {
          Optional<Expert> maybeFound =
              expertRepository.findById(
                  NumberUtils.isDigits(expertId) ? Integer.parseInt(expertId) : -1);
          if (maybeFound.isPresent()) manuallyFound = maybeFound.get();
          else {
            System.out.println("not found, try again");
            expertId = scanner.next();
          }
        }
        if (Objects.nonNull(manuallyFound)) {
          addToReport(
              report,
              true,
              "--- update expert {} {} {}",
              expertId,
              manuallyFound.getName(),
              manuallyFound.getEmail());
          updateExpertises(manuallyFound, expertNew, report);
        } else {
          addToReport(report, "--- create expert");
          expertRepository.save(buildExpertFromExpertNew(expertNew));
        }
      }
    }
    return report.toString();
  }
}
