package fr.gouv.beta.selexpert.service;

import fr.gouv.beta.selexpert.event.*;
import fr.gouv.beta.selexpert.model.Expert;
import fr.gouv.beta.selexpert.model.Request;
import fr.gouv.beta.selexpert.model.RequestExpert;
import fr.gouv.beta.selexpert.model.RequestExpertStatus;
import fr.gouv.beta.selexpert.model.RequestStatus;
import fr.gouv.beta.selexpert.model.User;
import fr.gouv.beta.selexpert.repository.RequestExpertRepository;
import fr.gouv.beta.selexpert.util.DateUtils;
import freemarker.template.Configuration;
import freemarker.template.TemplateException;
import jakarta.transaction.Transactional;
import java.io.IOException;
import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;
import lombok.Getter;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.collections4.ListUtils;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.context.event.EventListener;
import org.springframework.scheduling.annotation.Async;
import org.springframework.stereotype.Service;
import org.springframework.transaction.event.TransactionalEventListener;
import org.springframework.ui.freemarker.FreeMarkerTemplateUtils;
import sendinblue.ApiException;

@Service
@RequiredArgsConstructor
@Slf4j
public class MailService {

  private final Configuration freeMarkerConfiguration;
  private final RequestExpertRepository requestExpertRepository;
  private final MailSenderService mailSenderService;

  @Value("${selexpert.mail.sender-email}")
  private String senderEmail;

  @Value("${selexpert.mail.recipient-override}")
  private String recipientOverride;

  @Value("${selexpert.mail.activation-url}")
  private String activationUrl;

  @Value("${selexpert.mail.password-reset-url}")
  private String passwordResetUrl;

  @Value("${selexpert.url}")
  private String selexpertUrl;

  @Value("${selexpert.production}")
  private boolean isProduction;

  @Value("${selexpert.weekdays-to-reply}")
  private Integer weekdaysToReply;

  @Value(
      "#{T(java.time.format.DateTimeFormatter).ofPattern('${selexpert.date-format.deadline}').withLocale(T(java.util.Locale).FRANCE)}")
  private DateTimeFormatter dtfDeadline;

  @TransactionalEventListener
  @Transactional
  @Async
  public void onRequestCreated(RequestCreatedEvent requestCreatedEvent) {
    log.debug("RequestCreatedEvent");
    var request = requestCreatedEvent.request();
    sendExpertRequests(request);
    var emailModel = new HashMap<String, Object>();
    emailModel.put("procedureId", request.getProcedureId());
    emailModel.put("requestId", request.getId());
    if (CollectionUtils.isNotEmpty(request.getRequestClerks())) {
      request.getRequestClerks().stream()
          .filter( // all clerks except the one who created the request
              requestClerk ->
                  Objects.isNull(request.getClerk())
                      || !requestClerk.getId().equals(request.getClerk().getId()))
          .forEach(clerk -> send(clerk, Email.REQUEST_ASSIGNED, emailModel));
    }
  }

  @TransactionalEventListener
  @Transactional
  @Async
  public void onClerksAdded(ClerksAddedEvent clerksAddedEvent) {
    log.debug("ClerksAddedEvent");
    var request = clerksAddedEvent.request();
    var emailModel = new HashMap<String, Object>();
    emailModel.put("procedureId", request.getProcedureId());
    emailModel.put("requestId", request.getId());
    clerksAddedEvent
        .addedClerks()
        .forEach(clerk -> send(clerk, Email.REQUEST_ASSIGNED, emailModel));
  }

  @TransactionalEventListener
  @Transactional
  @Async
  public void onExpertsAdded(ExpertsAddedEvent expertsAddedEvent) {
    log.debug("ExpertsAddedEvent");
    var request = expertsAddedEvent.request();
    sendExpertRequests(request);
  }

  @TransactionalEventListener
  @Transactional
  @Async
  public void onRequestAccepted(RequestAcceptedEvent requestAcceptedEvent) {
    log.debug("RequestAcceptedEvent");
    var request = requestAcceptedEvent.request();
    var expertWhoAccepted = requestAcceptedEvent.expert();
    var emailModel = new HashMap<String, Object>();
    emailModel.put("expert", expertWhoAccepted.getName());
    emailModel.put("procedureId", request.getProcedureId());
    emailModel.put("requestId", request.getId());
    getCourtUsers(request)
        .forEach(courtUser -> send(courtUser, Email.REQUEST_ACCEPTED, emailModel));
    send(expertWhoAccepted, Email.REQUEST_ACCEPTED_ACK, emailModel);
    if (request.isUrgent()) {
      request.getRequestExperts().stream()
          .filter(re -> RequestExpertStatus.FULFILLED.equals(re.getStatus()))
          .forEach(
              requestExpert ->
                  send(requestExpert.getExpert(), Email.REQUEST_FULFILLED, emailModel));
    }
  }

  @TransactionalEventListener
  @Transactional
  @Async
  public void onRequestRefused(RequestRefusedEvent requestRefusedEvent) {
    log.debug("RequestRefusedEvent");
    var request = requestRefusedEvent.request();
    var expertWhoRefused = requestRefusedEvent.expert();
    sendRefusalEmailToCourtAndNextExpert(request, expertWhoRefused);
    var emailModel = new HashMap<String, Object>();
    emailModel.put("procedureId", request.getProcedureId());
    emailModel.put("requestId", request.getId());
    emailModel.put("reason", requestRefusedEvent.reason().getDisplayStr());
    emailModel.put("justification", requestRefusedEvent.justification());
    send(expertWhoRefused, Email.REQUEST_REFUSED_ACK, emailModel);
  }

  @TransactionalEventListener
  @Transactional
  @Async
  public void onRequestRefusedAuto(RequestRefusedAutoEvent requestRefusedAutoEvent) {
    log.debug("RequestRefusedAutoEvent");
    var request = requestRefusedAutoEvent.request();
    sendRefusalEmailToCourtAndNextExpert(
        request,
        !RequestStatus.REFUSED.equals(request.getStatus())
            ? requestRefusedAutoEvent.expertList().get(0)
            : null);
    Optional.ofNullable(requestRefusedAutoEvent.expertList())
        .orElse(List.of()) // should not happen
        .forEach(
            expert -> {
              var emailModel = new HashMap<String, Object>();
              emailModel.put("procedureId", request.getProcedureId());
              emailModel.put("requestId", request.getId());
              send(expert, Email.REQUEST_REFUSED_AUTO, emailModel);
            });
  }

  @EventListener
  @Async
  public void onDecisionAdded(DecisionAddedEvent decisionAddedEvent) {
    log.debug("DecisionAddedEvent");
    var request = decisionAddedEvent.request();
    var emailModel = new HashMap<String, Object>();
    emailModel.put("procedureId", request.getProcedureId());
    emailModel.put("requestId", request.getId());
    send(decisionAddedEvent.expertWhoAccepted(), Email.DECISION_ADDED, emailModel);
  }

  @EventListener
  @Async
  public void onRequestCancelled(RequestCancelledEvent requestCancelledEvent) {
    log.debug("RequestCancelledEvent");
    var request = requestCancelledEvent.request();
    var emailModel = new HashMap<String, Object>();
    emailModel.put("procedureId", request.getProcedureId());
    emailModel.put("requestId", request.getId());
    request.getRequestExperts().stream()
        .filter(
            re ->
                re.getStatus() == RequestExpertStatus.CANCELLED && Objects.nonNull(re.getSentOn()))
        .forEach(re -> send(re.getExpert(), Email.REQUEST_CANCELLED, emailModel));
  }

  @EventListener
  @Async
  public void onExpertActivatedEvent(ExpertActivatedEvent expertActivatedEvent) {
    log.debug("expertActivatedEvent");
    var email = expertActivatedEvent.expert().getEmail();
    var url = activationUrl + expertActivatedEvent.token();
    var emailModel = new HashMap<String, Object>();
    emailModel.put("url", url);
    emailModel.put("manuallyAddedExpert", expertActivatedEvent.expert().getManuallyAdded());
    log.info(url);
    send(email, Email.USER_ACTIVATION, emailModel);
    log.info("expert activation email sent to {}", email);
  }

  @EventListener
  @Async
  public void onCourtUserActivatedEvent(CourtUserActivatedEvent courtUserActivatedEvent) {
    log.debug("courtUserActivatedEvent");
    var email = courtUserActivatedEvent.user().getEmail();
    var url = passwordResetUrl + courtUserActivatedEvent.token();
    var emailModel = new HashMap<String, Object>();
    emailModel.put("url", url);
    log.info(url);
    send(email, Email.USER_ACTIVATION, emailModel);
    log.info("court user activation email sent to {}", email);
  }

  @EventListener
  @Async
  public void onPwdResetRequestedEvent(PwdResetRequestedEvent pwdResetRequestedEvent) {
    log.debug("pwdResetRequestedEvent");
    var email = pwdResetRequestedEvent.user().getEmail();
    var url = passwordResetUrl + pwdResetRequestedEvent.token();
    var emailModel = new HashMap<String, Object>();
    emailModel.put("url", url);
    log.info(url);
    send(email, Email.PASSWORD_RESET, emailModel);
    log.info("password reset email sent to {}", email);
  }

  @EventListener
  @Async
  public void onExpertIsNowAvailable(ExpertIsNowAvailableEvent expertIsNowAvailableEvent) {
    log.debug("expertIsNowAvailableEvent");
    send(expertIsNowAvailableEvent.expert(), Email.EXPERT_IS_NOW_AVAILABLE, new HashMap<>());
  }

  @EventListener
  @Async
  public void onExpertIsStillUnavailable(
      ExpertIsStillUnavailableEvent expertIsStillUnavailableEvent) {
    log.debug("expertIsStillUnavailableEvent");
    send(
        expertIsStillUnavailableEvent.expert(), Email.EXPERT_IS_STILL_UNAVAILABLE, new HashMap<>());
  }

  @EventListener
  @Async
  public void onRequestDeliberationReminder(
      RequestDeliberationReminderEvent requestDeliberationReminderEvent) {
    log.debug("RequestDeliberationReminderEvent");
    var request = requestDeliberationReminderEvent.request();
    var emailModel = new HashMap<String, Object>();
    emailModel.put("procedureId", request.getProcedureId());
    emailModel.put("requestId", request.getId());
    getCourtUsers(request)
        .forEach(courtUser -> send(courtUser, Email.REQUEST_DELIBERATION_REMINDER, emailModel));
  }

  @EventListener
  @Async
  public void onQuestionAsked(QuestionAskedEvent questionAskedEvent) {
    log.debug("QuestionAskedEvent");
    var request = questionAskedEvent.request();
    var expertWhoAsked = questionAskedEvent.expert();
    var emailModel = new HashMap<String, Object>();
    emailModel.put("expert", expertWhoAsked.getName());
    emailModel.put("procedureId", request.getProcedureId());
    emailModel.put("requestId", request.getId());
    getCourtUsers(request).forEach(courtUser -> send(courtUser, Email.QUESTION_ASKED, emailModel));
  }

  @EventListener
  @Async
  public void onQuestionAnswered(QuestionAnsweredEvent questionAnsweredEvent) {
    log.debug("QuestionAnsweredEvent");
    var request = questionAnsweredEvent.request();
    var expertWhoAsked = questionAnsweredEvent.expert();
    var emailModel = new HashMap<String, Object>();
    emailModel.put("procedureId", request.getProcedureId());
    emailModel.put("requestId", request.getId());
    send(expertWhoAsked, Email.QUESTION_ANSWERED, emailModel);
  }

  private void sendExpertRequests(Request request) {
    log.debug("Checking for emails to be sent for request " + request.getId());
    var now = LocalDateTime.now();
    List<RequestExpert> toSendList =
        requestExpertRepository.findByRequestAndStatus(request, RequestExpertStatus.SENDING);
    var emailModel = new HashMap<String, Object>();
    emailModel.put("procedureId", request.getProcedureId());
    emailModel.put("requestId", request.getId());
    emailModel.put("jurisdiction", request.getJurisdiction().getName());
    emailModel.put("deadline", dtfDeadline.format(DateUtils.getDeadline(now, weekdaysToReply)));
    emailModel.put("urgent", request.isUrgent());
    toSendList.forEach(
        toSend -> {
          var expert = toSend.getExpert();
          send(expert, Email.REQUEST_CREATED, emailModel);
          if (Objects.isNull(expert.getUserExpert())) { // non activated user, warn ourselves
            var emailToUsModel = new HashMap<String, Object>();
            emailToUsModel.put("request", request);
            emailToUsModel.put("expert", expert);
            send(senderEmail, Email.EXPERT_NOT_ACTIVATED, emailToUsModel);
          }
          toSend.updateStatus(RequestExpertStatus.SENT);
          toSend.setSentOn(now); // all requestExperts should have the same time
          // if urgent, possibly problematic if re1 sent at 23:59:59 and re2 at 00:00
        });
  }

  private void send(Expert expert, Email email, Map<String, Object> emailModel) {
    if (Objects.nonNull(expert.getUserExpert())) {
      send(expert.getUserExpert(), email, emailModel);
    } else if (StringUtils.isNotBlank(expert.getEmail())) {
      send(expert.getEmail(), email, emailModel);
    } else {
      log.warn("expert {} ({}) has no email address", expert.getName(), expert.getId());
    }
  }

  private void send(User user, Email email, Map<String, Object> emailModel) {
    if (email.optional && Boolean.FALSE.equals(user.getAllNotifications())) return;
    send(user.getEmail(), email, emailModel);
  }

  private void send(String to, Email email, Map<String, Object> emailModel) {
    log.info("sending email to: " + to);
    if (StringUtils.isBlank(to)) {
      log.warn("abort: no email address");
      return;
    }
    if (!isProduction) {
      emailModel.put("realRecipient", to);
      to = recipientOverride;
      log.info("recipientOverride: " + to);
    }
    emailModel.put("selexpertUrl", selexpertUrl);
    try {
      mailSenderService.send(
          to,
          email.subject,
          FreeMarkerTemplateUtils.processTemplateIntoString(
              freeMarkerConfiguration.getTemplate(email.template + ".ftlh"), emailModel));
    } catch (ApiException | IOException | TemplateException e) {
      log.error("error while sending email to: " + to, e);
    }
    log.info("email sent to: " + to);
  }

  private void sendRefusalEmailToCourtAndNextExpert(Request request, Expert expertWhoRefused) {
    var emailModel = new HashMap<String, Object>();
    emailModel.put("procedureId", request.getProcedureId());
    emailModel.put("requestId", request.getId());
    if (!RequestStatus.REFUSED.equals(request.getStatus())) {
      sendExpertRequests(request);
      emailModel.put("expert", expertWhoRefused.getName());
      emailModel.put("urgent", request.isUrgent());
      getCourtUsers(request)
          .forEach(courtUser -> send(courtUser, Email.REQUEST_REFUSED, emailModel));
    } else {
      getCourtUsers(request)
          .forEach(courtUser -> send(courtUser, Email.REQUEST_REFUSED_ALL, emailModel));
    }
  }

  private List<User> getCourtUsers(Request request) {
    return ListUtils.union(List.of(request.getMagistrate()), request.getRequestClerks());
  }

  @Getter
  public enum Email {

    // AUTH
    USER_ACTIVATION("Lien d'activation de compte", "user-activation"),
    PASSWORD_RESET("Réinitialisation de votre mot de passe", "password-reset"),

    // COURT (MAGISTRATE / CLERK)
    REQUEST_ACCEPTED("Demande acceptée", "request-accepted"),
    REQUEST_REFUSED("Un expert a refusé", "request-refused", true),
    REQUEST_REFUSED_ALL("Tous les experts ont refusé", "request-refused-all"),
    REQUEST_DELIBERATION_REMINDER(
        "La date du délibéré approche", "request-deliberation-reminder", true),
    QUESTION_ASKED("Une question a été posée", "question-asked"),
    // CLERK
    REQUEST_ASSIGNED("Rattachement à une demande d'expertise", "request-assigned", true),

    // EXPERT
    REQUEST_CREATED("Nouvelle demande d'expertise", "request-created"),
    REQUEST_ACCEPTED_ACK("Demande acceptée", "request-accepted-ack"),
    REQUEST_FULFILLED("Demande annulée", "request-fulfilled"),
    REQUEST_CANCELLED("Demande annulée", "request-cancelled"),
    REQUEST_REFUSED_AUTO("Refus automatique d'une demande", "request-refused-auto"),
    REQUEST_REFUSED_ACK("Demande refusée", "request-refused-ack"),
    DECISION_ADDED("Décision disponible", "decision-added"),
    EXPERT_IS_NOW_AVAILABLE("Votre disponibilité", "expert-is-now-available"),
    EXPERT_IS_STILL_UNAVAILABLE("Votre indisponibilité", "expert-is-still-unavailable"),
    QUESTION_ANSWERED("Réponse à votre question", "question-answered"),

    // ADMIN
    EXPERT_NOT_ACTIVATED(
        "ATTENTION : un expert non-activé a été sollicité", "expert-not-activated");

    private final String subject;
    private final String template;
    private final boolean optional;

    Email(String subject, String template, boolean optional) {
      this.subject = subject;
      this.template = template;
      this.optional = optional;
    }

    Email(String subject, String template) {
      this(subject, template, false);
    }
  }
}
