package fr.gouv.beta.selexpert;

import static org.assertj.core.api.BDDAssertions.then;

import fr.gouv.beta.selexpert.dto.ClerkDto;
import fr.gouv.beta.selexpert.dto.ExpertDto;
import fr.gouv.beta.selexpert.dto.LoginForm;
import fr.gouv.beta.selexpert.dto.MagistrateDto;
import fr.gouv.beta.selexpert.dto.PageOf;
import fr.gouv.beta.selexpert.dto.RequestCreation;
import fr.gouv.beta.selexpert.dto.RequestRefusalForm;
import fr.gouv.beta.selexpert.dto.UserInfo;
import fr.gouv.beta.selexpert.model.AppealCourt;
import fr.gouv.beta.selexpert.model.Expert;
import fr.gouv.beta.selexpert.model.Jurisdiction;
import fr.gouv.beta.selexpert.model.RefusalReason;
import fr.gouv.beta.selexpert.model.Request;
import fr.gouv.beta.selexpert.model.RequestExpertStatus;
import fr.gouv.beta.selexpert.model.RequestStatus;
import fr.gouv.beta.selexpert.model.Role;
import fr.gouv.beta.selexpert.model.UserClerk;
import fr.gouv.beta.selexpert.model.UserExpert;
import fr.gouv.beta.selexpert.model.UserMagistrate;
import fr.gouv.beta.selexpert.repository.AppealCourtRepository;
import fr.gouv.beta.selexpert.repository.ExpertRepository;
import fr.gouv.beta.selexpert.repository.JurisdictionRepository;
import fr.gouv.beta.selexpert.repository.RequestRepository;
import fr.gouv.beta.selexpert.repository.UserRepository;
import fr.gouv.beta.selexpert.service.MailSenderService;
import java.util.List;
import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.TestInstance;
import org.mockito.Mockito;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.boot.test.context.SpringBootTest.WebEnvironment;
import org.springframework.boot.test.context.TestConfiguration;
import org.springframework.boot.test.web.client.TestRestTemplate;
import org.springframework.boot.testcontainers.context.ImportTestcontainers;
import org.springframework.boot.testcontainers.service.connection.ServiceConnection;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Primary;
import org.springframework.core.ParameterizedTypeReference;
import org.springframework.http.HttpEntity;
import org.springframework.http.HttpHeaders;
import org.springframework.http.HttpMethod;
import org.springframework.security.crypto.bcrypt.BCryptPasswordEncoder;
import org.testcontainers.containers.PostgreSQLContainer;
import org.testcontainers.junit.jupiter.Container;
import org.testcontainers.junit.jupiter.Testcontainers;

@SpringBootTest(webEnvironment = WebEnvironment.DEFINED_PORT)
@Testcontainers
@ImportTestcontainers(SelexpertIT.Containers.class)
@TestInstance(TestInstance.Lifecycle.PER_CLASS)
class SelexpertIT {

  @Autowired AppealCourtRepository appealCourtRepository;
  @Autowired JurisdictionRepository jurisdictionRepository;
  @Autowired UserRepository userRepository;
  @Autowired ExpertRepository expertRepository;
  @Autowired RequestRepository requestRepository;

  @Autowired private TestRestTemplate rest;

  @Value("http://localhost:${server.port}/api")
  private String url;

  @BeforeAll
  void beforeAll() {
    var password = new BCryptPasswordEncoder().encode("pwd");
    AppealCourt appealCourt = AppealCourt.builder().name("CA").code("0").build();
    appealCourtRepository.save(appealCourt);
    Jurisdiction jurisdiction = Jurisdiction.builder().name("J").appealCourt(appealCourt).build();
    jurisdictionRepository.save(jurisdiction);
    Expert e1 = Expert.builder().name("e1").email("e1@expert.fr").appealCourt(appealCourt).build();
    expertRepository.save(e1);
    userRepository.save(
        UserExpert.builder()
            .name("e1")
            .email("e1@expert.fr")
            .password(password)
            .role(Role.EXPERT)
            .expert(e1)
            .build());
    Expert e2 = Expert.builder().name("e2").email("e2@expert.fr").appealCourt(appealCourt).build();
    expertRepository.save(e2);
    userRepository.save(
        UserExpert.builder()
            .name("e2")
            .email("e2@expert.fr")
            .password(password)
            .role(Role.EXPERT)
            .expert(e2)
            .build());
    userRepository.save(
        UserMagistrate.builder()
            .name("m1")
            .email("m1@justice.fr")
            .password(password)
            .role(Role.MAGISTRATE)
            .jurisdiction(jurisdiction)
            .build());
    userRepository.save(
        UserClerk.builder()
            .name("g1")
            .email("g1@justice.fr")
            .password(password)
            .role(Role.CLERK)
            .jurisdiction(jurisdiction)
            .build());
  }

  @Test
  void requestCreationFlow() {

    // connect as clerk
    var authHeader = getAuthHeader("g1@justice.fr");

    // get expert list
    var respExpertList =
        rest.exchange(
            url + "/court/experts?page=0",
            HttpMethod.GET,
            new HttpEntity<>(authHeader),
            new ParameterizedTypeReference<PageOf<ExpertDto>>() {});
    var expertList = respExpertList.getBody().getElements();
    then(expertList.size()).isEqualTo(2);

    // get magistrate list
    var respMagistrateList =
        rest.exchange(
            url + "/court/magistrates",
            HttpMethod.GET,
            new HttpEntity<>(authHeader),
            new ParameterizedTypeReference<List<MagistrateDto>>() {});
    var magistrateList = respMagistrateList.getBody();
    then(magistrateList.size()).isEqualTo(1);

    // get clerk list
    var respClerkList =
        rest.exchange(
            url + "/court/clerks",
            HttpMethod.GET,
            new HttpEntity<>(authHeader),
            new ParameterizedTypeReference<List<ClerkDto>>() {});
    var clerkList = respClerkList.getBody();
    then(clerkList.size()).isEqualTo(1);

    // create a request
    RequestCreation requestCreation =
        RequestCreation.builder()
            .expertIds(expertList.stream().map(ExpertDto::getId).toList())
            .magistrateId(magistrateList.get(0).getId())
            .clerkIds(clerkList.stream().map(ClerkDto::getId).toList())
            .procedureId("123")
            .extraInfo("info")
            .urgent(false)
            .build();
    var respRequest =
        rest.exchange(
            url + "/court/request",
            HttpMethod.POST,
            new HttpEntity<>(requestCreation, authHeader),
            new ParameterizedTypeReference<Integer>() {});
    var requestId = respRequest.getBody();

    // check request
    Request request = requestRepository.findById(requestId).orElseThrow();
    then(request.getStatus()).isEqualTo(RequestStatus.OPEN);
    then(request.getRequestExperts().size()).isEqualTo(2);
    then(request.getRequestExperts().get(0).getStatus()).isEqualTo(RequestExpertStatus.SENDING);
    then(request.getRequestExperts().get(1).getStatus()).isEqualTo(RequestExpertStatus.QUEUED);

    // refuse request
    var expertRefuses = request.getRequestExperts().get(0).getExpert();
    RequestRefusalForm requestRefusalForm =
        RequestRefusalForm.builder().reason(RefusalReason.INCOMPATIBLE).build();
    authHeader = getAuthHeader(expertRefuses.getEmail());
    var respRefuseRequest =
        rest.exchange(
            url + "/expert/request/" + requestId + "/refuse",
            HttpMethod.POST,
            new HttpEntity<>(requestRefusalForm, authHeader),
            new ParameterizedTypeReference<Void>() {});

    // check request
    request = requestRepository.findById(requestId).orElseThrow();
    then(request.getStatus()).isEqualTo(RequestStatus.OPEN);
    then(request.getRequestExperts().get(0).getStatus()).isEqualTo(RequestExpertStatus.REFUSED);
    then(request.getRequestExperts().get(1).getStatus()).isEqualTo(RequestExpertStatus.SENDING);

    // accept request
    var expertAccepts = request.getRequestExperts().get(1).getExpert();
    authHeader = getAuthHeader(expertAccepts.getEmail());
    var respAcceptRequest =
        rest.exchange(
            url + "/expert/request/" + requestId + "/accept",
            HttpMethod.POST,
            new HttpEntity<>(authHeader),
            new ParameterizedTypeReference<Void>() {});

    // check request
    request = requestRepository.findById(requestId).orElseThrow();
    then(request.getStatus()).isEqualTo(RequestStatus.ACCEPTED);
  }

  private HttpHeaders getAuthHeader(String email) {
    UserInfo userInfo =
        rest.postForObject(url + "/auth/login", new LoginForm(email, "pwd"), UserInfo.class);
    HttpHeaders headers = new HttpHeaders();
    headers.add("Authorization", "Bearer " + userInfo.getAccessToken());
    return headers;
  }

  @Testcontainers
  interface Containers {
    // https://stackoverflow.com/questions/61625288/spring-boot-testcontainers-mapped-port-can-only-be-obtained-after-the-container/78168562#78168562
    @Container @ServiceConnection
    PostgreSQLContainer<?> postgres = new PostgreSQLContainer<>("postgres:14-alpine");
  }

  @TestConfiguration
  static class IntegrationTestConfig {
    @Bean
    @Primary
    public MailSenderService mailSenderService() {
      MailSenderService mailSenderService = Mockito.mock(MailSenderService.class);
      // Mockito.doReturn(XXX).when(XXX).getXXX();
      return mailSenderService;
    }
  }
}
