package fr.gouv.beta.selexpert.service;

import static org.assertj.core.api.BDDAssertions.catchThrowable;
import static org.assertj.core.api.BDDAssertions.then;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.BDDMockito.given;
import static org.mockito.Mockito.never;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;

import fr.gouv.beta.selexpert.AppException;
import fr.gouv.beta.selexpert.model.Expert;
import fr.gouv.beta.selexpert.model.File;
import fr.gouv.beta.selexpert.model.FileType;
import fr.gouv.beta.selexpert.model.Request;
import fr.gouv.beta.selexpert.model.RequestExpert;
import fr.gouv.beta.selexpert.model.RequestExpertStatus;
import fr.gouv.beta.selexpert.model.Role;
import fr.gouv.beta.selexpert.model.User;
import fr.gouv.beta.selexpert.model.UserClerk;
import fr.gouv.beta.selexpert.model.UserExpert;
import fr.gouv.beta.selexpert.model.UserMagistrate;
import fr.gouv.beta.selexpert.repository.FileRepository;
import fr.gouv.beta.selexpert.repository.RequestRepository;
import fr.gouv.beta.selexpert.repository.UserRepository;
import java.util.List;
import java.util.Optional;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.extension.ExtendWith;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.EnumSource;
import org.junit.jupiter.params.provider.EnumSource.Mode;
import org.junit.jupiter.params.provider.ValueSource;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.junit.jupiter.MockitoExtension;
import org.springframework.test.util.ReflectionTestUtils;

@ExtendWith(MockitoExtension.class)
class CommonServiceTest {

  @Mock private RequestRepository requestRepository;
  @Mock private FileService fileService;
  @Mock private FileRepository fileRepository;
  @Mock private UserRepository userRepository;

  @InjectMocks private CommonService commonService;

  private static Request makeRequest(Role role) {
    Request request = new Request();
    switch (role) {
      case MAGISTRATE -> request.setMagistrate(
          UserMagistrate.builder().role(Role.MAGISTRATE).id(1).build());
      case CLERK -> request.setRequestClerks(
          List.of(UserClerk.builder().role(Role.CLERK).id(1).build()));
      case EXPERT -> request.setRequestExperts(
          List.of(
              new RequestExpert(
                  request,
                  Expert.builder()
                      .userExpert(UserExpert.builder().role(Role.EXPERT).id(1).build())
                      .build(),
                  1)));
    }
    return request;
  }

  @BeforeEach
  public void setUp() {
    ReflectionTestUtils.setField(commonService, "monthsToArchiveRequest", 4);
  }

  @ParameterizedTest
  @EnumSource(
      value = Role.class,
      mode = Mode.EXCLUDE,
      names = {"ADMIN"})
  public void download(Role role) {

    // GIVEN
    Request request = makeRequest(role);
    User user = new User();
    user.setId(1);
    user.setRole(role);
    given(fileRepository.findByName("file"))
        .willReturn(Optional.of(File.builder().request(request).type(FileType.ATTACHMENT).build()));

    // WHEN
    commonService.download("file", user);

    // THEN
    verify(fileService).download("file");
  }

  @ParameterizedTest
  @EnumSource(value = Role.class)
  public void download_notAllowed(Role role) {

    // GIVEN
    Request request = makeRequest(role);
    User user = new User();
    user.setId(2);
    user.setRole(role);
    given(fileRepository.findByName("file"))
        .willReturn(Optional.of(File.builder().request(request).type(FileType.ATTACHMENT).build()));

    // WHEN
    Throwable throwable = catchThrowable(() -> commonService.download("file", user));

    // THEN
    then(throwable).isInstanceOf(AppException.class).hasMessageContaining("can't be accessed");
    verify(fileService, never()).download(any());
  }

  @Test
  public void download_decision() {

    // GIVEN
    Request request = makeRequest(Role.EXPERT);
    request.getRequestExperts().get(0).setStatus(RequestExpertStatus.ACCEPTED);
    given(fileRepository.findByName("file"))
        .willReturn(Optional.of(File.builder().request(request).type(FileType.DECISION).build()));

    // WHEN
    commonService.download("file", request.getRequestExperts().get(0).getExpert().getUserExpert());

    // THEN
    verify(fileService).download("file");
  }

  @Test
  public void download_decision_notAllowed() {

    // GIVEN
    Request request = makeRequest(Role.EXPERT);
    request.getRequestExperts().get(0).setStatus(RequestExpertStatus.REFUSED);
    given(fileRepository.findByName("file"))
        .willReturn(Optional.of(File.builder().request(request).type(FileType.DECISION).build()));

    // WHEN
    Throwable throwable =
        catchThrowable(
            () ->
                commonService.download(
                    "file", request.getRequestExperts().get(0).getExpert().getUserExpert()));

    // THEN
    then(throwable).isInstanceOf(AppException.class).hasMessageContaining("can't be accessed");
    verify(fileService, never()).download(any());
  }

  @Test
  public void archiveRequests() {

    // GIVEN
    Request request = Request.builder().files(List.of(new File(), new File())).build();
    given(requestRepository.getRequestsToBeArchived(any())).willReturn(List.of(request));

    // WHEN
    commonService.archiveRequests();

    // THEN
    then(request.isArchived()).isTrue();
    verify(fileService, times(request.getFiles().size())).delete(any());
  }

  @ParameterizedTest
  @ValueSource(strings = {"true", "false", "null"})
  public void deleteFile(String allNotificationsStr) {

    // GIVEN
    Boolean allNotifications =
        "null".equals(allNotificationsStr) ? null : Boolean.parseBoolean(allNotificationsStr);
    User user = User.builder().allNotifications(allNotifications).build();
    given(userRepository.findById(1)).willReturn(Optional.of(user));

    // WHEN
    commonService.toggleNotifications(1);

    // THEN
    then(user.getAllNotifications()).isEqualTo(Boolean.FALSE.equals(allNotifications));
  }
}
