package fr.gouv.beta.selexpert.service;

import static org.assertj.core.api.BDDAssertions.catchThrowable;
import static org.assertj.core.api.BDDAssertions.then;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.BDDMockito.given;
import static org.mockito.Mockito.never;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;

import fr.gouv.beta.selexpert.AppException;
import fr.gouv.beta.selexpert.dto.AddExpertsToRequest;
import fr.gouv.beta.selexpert.dto.AddFile;
import fr.gouv.beta.selexpert.dto.AddNote;
import fr.gouv.beta.selexpert.dto.CancelRequest;
import fr.gouv.beta.selexpert.dto.DeleteFile;
import fr.gouv.beta.selexpert.dto.DeliberationDateUpdate;
import fr.gouv.beta.selexpert.dto.RequestCreation;
import fr.gouv.beta.selexpert.dto.RequestCreation.RequestCreationBuilder;
import fr.gouv.beta.selexpert.dto.UpdateRequestClerks;
import fr.gouv.beta.selexpert.event.ClerksAddedEvent;
import fr.gouv.beta.selexpert.event.DecisionAddedEvent;
import fr.gouv.beta.selexpert.event.ExpertsAddedEvent;
import fr.gouv.beta.selexpert.event.QuestionAnsweredEvent;
import fr.gouv.beta.selexpert.event.RequestCancelledEvent;
import fr.gouv.beta.selexpert.event.RequestCreatedEvent;
import fr.gouv.beta.selexpert.event.RequestDeliberationReminderEvent;
import fr.gouv.beta.selexpert.model.AppealCourt;
import fr.gouv.beta.selexpert.model.Expert;
import fr.gouv.beta.selexpert.model.File;
import fr.gouv.beta.selexpert.model.FileType;
import fr.gouv.beta.selexpert.model.Jurisdiction;
import fr.gouv.beta.selexpert.model.Note;
import fr.gouv.beta.selexpert.model.NoteType;
import fr.gouv.beta.selexpert.model.Request;
import fr.gouv.beta.selexpert.model.Request.RequestBuilder;
import fr.gouv.beta.selexpert.model.RequestExpert;
import fr.gouv.beta.selexpert.model.RequestExpertStatus;
import fr.gouv.beta.selexpert.model.RequestStatus;
import fr.gouv.beta.selexpert.model.Role;
import fr.gouv.beta.selexpert.model.User;
import fr.gouv.beta.selexpert.model.UserClerk;
import fr.gouv.beta.selexpert.model.UserExpert;
import fr.gouv.beta.selexpert.model.UserMagistrate;
import fr.gouv.beta.selexpert.repository.ExpertRepository;
import fr.gouv.beta.selexpert.repository.FileRepository;
import fr.gouv.beta.selexpert.repository.NoteRepository;
import fr.gouv.beta.selexpert.repository.RequestRepository;
import fr.gouv.beta.selexpert.repository.UserClerkRepository;
import java.time.LocalDate;
import java.time.ZoneId;
import java.util.ArrayList;
import java.util.List;
import java.util.Optional;
import java.util.function.Function;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.extension.ExtendWith;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.EnumSource;
import org.junit.jupiter.params.provider.ValueSource;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.invocation.InvocationOnMock;
import org.mockito.junit.jupiter.MockitoExtension;
import org.springframework.context.ApplicationEventPublisher;
import org.springframework.mock.web.MockMultipartFile;
import org.springframework.test.util.ReflectionTestUtils;
import org.springframework.web.multipart.MultipartFile;

@ExtendWith(MockitoExtension.class)
class CourtServiceTest {

  // simulate db save (add id)
  private final Function<InvocationOnMock, Request> requestSaveFn =
      invocation -> {
        var request = (Request) invocation.getArgument(0);
        request.setId(1);
        return request;
      };
  @Mock private ExpertRepository expertRepository;
  @Mock private UserClerkRepository userClerkRepository;
  @Mock private RequestRepository requestRepository;
  @Mock private ApplicationEventPublisher eventPublisher;
  @Mock private FileService fileService;
  @Mock private FileRepository fileRepository;
  @Mock private NoteRepository noteRepository;

  @InjectMocks private CourtService courtService;

  private static RequestCreationBuilder makeRequestCreation() {
    return RequestCreation.builder()
        .procedureId("123")
        .extraInfo("extra info")
        .urgent(false)
        .clerkIds(List.of(1, 2))
        .expertIds(List.of(1, 2));
  }

  private static RequestBuilder makeRequest(RequestStatus status, boolean isUrgent) {
    return Request.builder()
        .id(1)
        .jurisdiction(makeJurisdiction())
        .status(status)
        .urgent(isUrgent)
        .files(
            new ArrayList<>(List.of(File.builder().name("1/attachment-4.pdf").ranking(4).build())))
        .requestExperts(new ArrayList<>(List.of(new RequestExpert(), new RequestExpert())));
  }

  private static UserMagistrate makeMagistrate() {
    return UserMagistrate.builder().role(Role.MAGISTRATE).jurisdiction(makeJurisdiction()).build();
  }

  private static UserClerk makeClerk(Integer id) {
    return UserClerk.builder().role(Role.CLERK).jurisdiction(makeJurisdiction()).id(id).build();
  }

  private static AppealCourt makeAppealCourt() {
    return AppealCourt.builder().id(1).build();
  }

  private static Jurisdiction makeJurisdiction() {
    return Jurisdiction.builder().id(1).appealCourt(makeAppealCourt()).build();
  }

  private static User makeAdmin() {
    return User.builder().role(Role.ADMIN).build();
  }

  @BeforeEach
  public void setUp() {
    ReflectionTestUtils.setField(courtService, "daysToRemindDeliberation", 7);
  }

  @ParameterizedTest
  @ValueSource(strings = {"false", "true"})
  public void createRequest(String urgentStr) {

    // GIVEN
    boolean urgent = Boolean.getBoolean(urgentStr);
    RequestCreation requestCreation = makeRequestCreation().urgent(urgent).build();
    UserMagistrate magistrate = UserMagistrate.builder().build();
    UserClerk clerk = UserClerk.builder().build();
    given(expertRepository.findById(any()))
        .willReturn(Optional.of(Expert.builder().email("@").build()));
    given(userClerkRepository.findByIdAndJurisdiction(any(), any()))
        .willReturn(Optional.of(new UserClerk()));
    given(requestRepository.save(any())).willAnswer(requestSaveFn::apply);

    // WHEN
    Request request = courtService.createRequest(requestCreation, magistrate, clerk);

    // THEN
    then(request).isNotNull();
    then(request.getJurisdiction()).isEqualTo(magistrate.getJurisdiction());
    then(request.getMagistrate()).isEqualTo(magistrate);
    then(request.getClerk()).isEqualTo(clerk);
    then(request.getProcedureId()).isEqualTo(requestCreation.getProcedureId());
    then(request.getExtraInfo()).isEqualTo(requestCreation.getExtraInfo());
    then(request.isUrgent()).isEqualTo(requestCreation.getUrgent());
    then(request.getRequestExperts().size()).isEqualTo(requestCreation.getExpertIds().size());
    then(request.getRequestClerks().size()).isEqualTo(requestCreation.getClerkIds().size() + 1);
    var sendingRequestExperts =
        request.getRequestExperts().stream()
            .filter(re -> RequestExpertStatus.SENDING.equals(re.getStatus()))
            .toList();
    then(sendingRequestExperts.size()).isEqualTo(urgent ? request.getRequestExperts().size() : 1);
    if (!urgent) then(sendingRequestExperts.get(0).getRanking()).isEqualTo(1);
    verify(eventPublisher).publishEvent(new RequestCreatedEvent(request));
  }

  @Test
  public void createRequest_no_experts() {

    // GIVEN
    RequestCreation requestCreation = makeRequestCreation().expertIds(null).build();

    // WHEN
    Throwable throwable =
        catchThrowable(() -> courtService.createRequest(requestCreation, makeMagistrate(), null));

    // THEN
    then(throwable).isInstanceOf(AppException.class).hasMessageContaining("no experts");
  }

  @Test
  public void createRequest_expert_not_found() {

    // GIVEN
    RequestCreation requestCreation = makeRequestCreation().build();
    given(expertRepository.findById(any())).willReturn(Optional.empty());

    // WHEN
    Throwable throwable =
        catchThrowable(() -> courtService.createRequest(requestCreation, makeMagistrate(), null));

    // THEN
    then(throwable).isInstanceOf(AppException.class).hasMessageContaining("expert not found");
  }

  @Test
  public void createRequest_expert_no_email() {

    // GIVEN
    RequestCreation requestCreation = makeRequestCreation().build();
    UserMagistrate magistrate = UserMagistrate.builder().build();
    given(expertRepository.findById(any())).willReturn(Optional.of(new Expert()));

    // WHEN
    Throwable throwable =
        catchThrowable(() -> courtService.createRequest(requestCreation, magistrate, null));

    // THEN
    then(throwable)
        .isInstanceOf(AppException.class)
        .hasMessageContaining("can't select this expert (no email)");
  }

  @Test
  public void createRequest_clerk_not_found() {

    // GIVEN
    RequestCreation requestCreation = makeRequestCreation().build();
    UserMagistrate magistrate = UserMagistrate.builder().build();
    given(expertRepository.findById(any()))
        .willReturn(Optional.of(Expert.builder().email("@").build()));
    given(userClerkRepository.findByIdAndJurisdiction(any(), any())).willReturn(Optional.empty());

    // WHEN
    Throwable throwable =
        catchThrowable(() -> courtService.createRequest(requestCreation, magistrate, null));

    // THEN
    then(throwable).isInstanceOf(AppException.class).hasMessageContaining("clerk not found");
  }

  @ParameterizedTest
  @ValueSource(strings = {"requestOpen", "requestRefusedNonUrgent", "requestRefusedUrgent"})
  public void addExpertsToRequest(String scenario) {

    // GIVEN
    Request request;
    switch (scenario) {
      case "requestOpen" -> request = makeRequest(RequestStatus.OPEN, false).build();
      case "requestRefusedNonUrgent" -> request = makeRequest(RequestStatus.REFUSED, false).build();
      case "requestRefusedUrgent" -> request = makeRequest(RequestStatus.REFUSED, true).build();
      default -> throw new RuntimeException("scenario not found");
    }
    UserMagistrate magistrate = makeMagistrate();
    AddExpertsToRequest addExpertsToRequest = new AddExpertsToRequest(1, List.of(3, 4));
    given(
            requestRepository.findByIdAndJurisdictionAndMagistrate(
                1, magistrate.getJurisdiction(), magistrate))
        .willReturn(Optional.of(request));
    given(expertRepository.findById(any()))
        .willReturn(Optional.of(Expert.builder().email("@").build()));

    // WHEN
    courtService.addExpertsToRequest(addExpertsToRequest, magistrate);

    // THEN
    then(request.getRequestExperts().size()).isEqualTo(4);
    then(request.getStatus()).isEqualTo(RequestStatus.OPEN);
    switch (scenario) {
      case "requestOpen" -> {
        then(request.getRequestExperts().get(2).getStatus()).isEqualTo(RequestExpertStatus.QUEUED);
        then(request.getRequestExperts().get(3).getStatus()).isEqualTo(RequestExpertStatus.QUEUED);
      }
      case "requestRefusedNonUrgent" -> {
        then(request.getRequestExperts().get(2).getStatus()).isEqualTo(RequestExpertStatus.SENDING);
        then(request.getRequestExperts().get(3).getStatus()).isEqualTo(RequestExpertStatus.QUEUED);
      }
      case "requestRefusedUrgent" -> {
        then(request.getRequestExperts().get(2).getStatus()).isEqualTo(RequestExpertStatus.SENDING);
        then(request.getRequestExperts().get(3).getStatus()).isEqualTo(RequestExpertStatus.SENDING);
      }
      default -> throw new RuntimeException("scenario not found");
    }
    verify(eventPublisher).publishEvent(new ExpertsAddedEvent(request));
  }

  @ParameterizedTest
  @EnumSource(
      value = RequestStatus.class,
      names = {"OPEN", "REFUSED"},
      mode = EnumSource.Mode.EXCLUDE)
  public void addExpertsToRequest_status_not_allowed(RequestStatus status) {

    // GIVEN
    Request request = makeRequest(status, false).build();
    UserMagistrate magistrate = makeMagistrate();
    AddExpertsToRequest addExpertsToRequest = new AddExpertsToRequest(1, List.of(3));
    given(
            requestRepository.findByIdAndJurisdictionAndMagistrate(
                1, magistrate.getJurisdiction(), magistrate))
        .willReturn(Optional.of(request));

    // WHEN
    Throwable throwable =
        catchThrowable(() -> courtService.addExpertsToRequest(addExpertsToRequest, magistrate));

    // THEN
    then(throwable).isInstanceOf(AppException.class).hasMessageContaining("can't update request");
  }

  @Test
  public void addExpertsToRequest_archived_not_allowed() {

    // GIVEN
    Request request = makeRequest(RequestStatus.REFUSED, false).archived(true).build();
    UserMagistrate magistrate = makeMagistrate();
    AddExpertsToRequest addExpertsToRequest = new AddExpertsToRequest(1, List.of(3));
    given(
            requestRepository.findByIdAndJurisdictionAndMagistrate(
                1, magistrate.getJurisdiction(), magistrate))
        .willReturn(Optional.of(request));

    // WHEN
    Throwable throwable =
        catchThrowable(() -> courtService.addExpertsToRequest(addExpertsToRequest, magistrate));

    // THEN
    then(throwable).isInstanceOf(AppException.class).hasMessageContaining("can't update request");
  }

  @Test
  public void addFile_attachment() {

    // GIVEN
    AddFile addFile = new AddFile(1, FileType.ATTACHMENT, "");
    MultipartFile multipartFile = new MockMultipartFile("file", (byte[]) null);
    UserClerk clerk = makeClerk(1);
    Request request = makeRequest(RequestStatus.OPEN, false).build();
    request.getFiles().add(File.builder().ranking(4).build());
    given(
            requestRepository.findByIdAndJurisdictionAndRequestClerks(
                1, clerk.getJurisdiction(), clerk))
        .willReturn(Optional.of(request));

    // WHEN
    String fileName = courtService.addFile(addFile, multipartFile, clerk);

    // THEN
    then(fileName).isEqualTo("1/attachment-5.pdf");
    verify(fileService).upload(multipartFile, fileName);
    verify(fileRepository).save(any());
  }

  @Test
  public void addFile_decision() {

    // GIVEN
    AddFile addFile = new AddFile(1, FileType.DECISION, "");
    MultipartFile multipartFile = new MockMultipartFile("file", (byte[]) null);
    UserClerk clerk = makeClerk(1);
    Request request = makeRequest(RequestStatus.ACCEPTED, false).build();
    var requestExpertAccepted = request.getRequestExperts().get(0);
    requestExpertAccepted.setStatus(RequestExpertStatus.ACCEPTED);
    requestExpertAccepted.setExpert(new Expert());
    given(
            requestRepository.findByIdAndJurisdictionAndRequestClerks(
                1, clerk.getJurisdiction(), clerk))
        .willReturn(Optional.of(request));

    // WHEN
    String fileName = courtService.addFile(addFile, multipartFile, clerk);

    // THEN
    then(fileName).isEqualTo("1/decision.pdf");
    verify(fileService).upload(multipartFile, fileName);
    verify(fileRepository).save(any());
    verify(eventPublisher)
        .publishEvent(new DecisionAddedEvent(request, requestExpertAccepted.getExpert()));
  }

  @ParameterizedTest
  @ValueSource(ints = {4, 5})
  public void deleteFile(int attachmentNumber) {

    // GIVEN
    DeleteFile deleteFile = new DeleteFile("1/attachment-" + attachmentNumber + ".pdf");
    User admin = makeAdmin();
    Request request = makeRequest(RequestStatus.OPEN, false).build();
    given(requestRepository.findById(1)).willReturn(Optional.of(request));

    // WHEN
    courtService.deleteFile(deleteFile, admin);

    // THEN
    verify(fileService, times(attachmentNumber == 4 ? 1 : 0)).delete(deleteFile.getFileName());
  }

  @Test
  public void cancelRequest() {

    // GIVEN
    CancelRequest cancelRequest = new CancelRequest();
    cancelRequest.setRequestId(1);
    Request request = makeRequest(RequestStatus.OPEN, false).build();
    RequestExpert requestExpert1 =
        RequestExpert.builder().status(RequestExpertStatus.ACCEPTED).build();
    RequestExpert requestExpert2 = RequestExpert.builder().status(RequestExpertStatus.SENT).build();
    RequestExpert requestExpert3 =
        RequestExpert.builder().status(RequestExpertStatus.QUEUED).build();
    request.setRequestExperts(List.of(requestExpert1, requestExpert2, requestExpert3));
    given(requestRepository.findById(1)).willReturn(Optional.of(request));

    // WHEN
    courtService.cancelRequest(cancelRequest, makeAdmin());

    // THEN
    then(request.getStatus()).isEqualTo(RequestStatus.CANCELLED);
    then(requestExpert1.getStatus()).isEqualTo(RequestExpertStatus.ACCEPTED);
    then(requestExpert2.getStatus()).isEqualTo(RequestExpertStatus.CANCELLED);
    then(requestExpert3.getStatus()).isEqualTo(RequestExpertStatus.CANCELLED);
    verify(eventPublisher).publishEvent(new RequestCancelledEvent(request));
  }

  @ParameterizedTest
  @ValueSource(ints = {6, 7, 8})
  public void updateDeliberationDate(int daysFromNow) {

    // GIVEN
    var deliberationDate = LocalDate.now(ZoneId.of("Europe/Paris")).plusDays(daysFromNow);
    DeliberationDateUpdate deliberationDateUpdate = new DeliberationDateUpdate();
    deliberationDateUpdate.setRequestId(1);
    deliberationDateUpdate.setDeliberationDate(deliberationDate);
    Request request = makeRequest(RequestStatus.OPEN, false).build();
    given(requestRepository.findById(1)).willReturn(Optional.of(request));

    // WHEN
    courtService.updateDeliberationDate(deliberationDateUpdate, makeAdmin());

    // THEN
    then(request.getDeliberationDate()).isEqualTo(deliberationDate);
    verify(eventPublisher, times(daysFromNow <= 7 ? 1 : 0))
        .publishEvent(new RequestDeliberationReminderEvent(request));
  }

  @Test
  public void updateClerks() {

    // GIVEN
    UpdateRequestClerks updateRequestClerks = new UpdateRequestClerks(1, List.of(1, 3, 4));
    Request request = makeRequest(RequestStatus.OPEN, false).build();
    request.setMagistrate(makeMagistrate());
    request.getMagistrate().setJurisdiction(new Jurisdiction());
    request.setRequestClerks(new ArrayList<>(List.of(makeClerk(1), makeClerk(2))));
    UserClerk addedClerk = makeClerk(3);
    given(requestRepository.findById(1)).willReturn(Optional.of(request));
    given(userClerkRepository.findByIdAndJurisdiction(3, request.getMagistrate().getJurisdiction()))
        .willReturn(Optional.of(addedClerk));
    given(userClerkRepository.findByIdAndJurisdiction(4, request.getMagistrate().getJurisdiction()))
        .willReturn(Optional.empty());

    // WHEN
    courtService.updateClerks(updateRequestClerks, makeAdmin());

    // THEN
    then(request.getRequestClerks().stream().map(User::getId).toList()).isEqualTo(List.of(1, 3));
    verify(eventPublisher).publishEvent(new ClerksAddedEvent(request, List.of(addedClerk)));
  }

  @Test
  public void addNote_internal_note() {

    // GIVEN
    AddNote addNote = AddNote.builder().requestId(1).content("hello").build();
    Request request = makeRequest(RequestStatus.OPEN, false).build();
    UserMagistrate userMagistrate = makeMagistrate();
    given(
            requestRepository.findByIdAndJurisdictionAndMagistrate(
                addNote.getRequestId(), userMagistrate.getJurisdiction(), userMagistrate))
        .willReturn(Optional.of(request));

    // WHEN
    Note internalNote = courtService.addNote(addNote, userMagistrate);

    // THEN
    verify(noteRepository).save(any());
    then(internalNote.getContent()).isEqualTo(addNote.getContent());
    then(internalNote.getRequest()).isEqualTo(request);
    then(internalNote.getType()).isEqualTo(NoteType.INTERNAL);
    then(internalNote.getUser()).isEqualTo(userMagistrate);
    then(internalNote.getQuestionId()).isNull();
    verify(eventPublisher, never()).publishEvent(any());
  }

  @Test
  public void addNote_answer() {

    // GIVEN
    AddNote addNote = AddNote.builder().requestId(1).content("hello").questionId(3).build();
    Request request = makeRequest(RequestStatus.OPEN, false).build();
    Note question =
        Note.builder()
            .type(NoteType.QUESTION)
            .id(addNote.getQuestionId())
            .user(UserExpert.builder().expert(new Expert()).build())
            .build();
    request.getNotes().add(question);
    UserMagistrate userMagistrate = makeMagistrate();
    given(
            requestRepository.findByIdAndJurisdictionAndMagistrate(
                addNote.getRequestId(), userMagistrate.getJurisdiction(), userMagistrate))
        .willReturn(Optional.of(request));

    // WHEN
    Note internalNote = courtService.addNote(addNote, userMagistrate);

    // THEN
    verify(noteRepository).save(any());
    then(internalNote.getContent()).isEqualTo(addNote.getContent());
    then(internalNote.getRequest()).isEqualTo(request);
    then(internalNote.getType()).isEqualTo(NoteType.ANSWER);
    then(internalNote.getUser()).isEqualTo(userMagistrate);
    then(internalNote.getQuestionId()).isEqualTo(question.getId());
    verify(eventPublisher)
        .publishEvent(
            new QuestionAnsweredEvent(request, ((UserExpert) question.getUser()).getExpert()));
  }

  @Test
  public void addNote_answer_no_matching_question() {

    // GIVEN
    AddNote addNote = AddNote.builder().requestId(1).content("hello").questionId(3).build();
    Request request = makeRequest(RequestStatus.OPEN, false).build();
    UserMagistrate userMagistrate = makeMagistrate();
    given(
            requestRepository.findByIdAndJurisdictionAndMagistrate(
                addNote.getRequestId(), userMagistrate.getJurisdiction(), userMagistrate))
        .willReturn(Optional.of(request));

    // WHEN
    Throwable throwable = catchThrowable(() -> courtService.addNote(addNote, userMagistrate));

    // THEN
    then(throwable)
        .isInstanceOf(AppException.class)
        .hasMessageContaining("question not found in this request");
  }

  @Test
  public void getRequest() {

    // GIVEN
    Request request = makeRequest(RequestStatus.OPEN, false).build();
    UserMagistrate userMagistrate = makeMagistrate();
    given(
            requestRepository.findByIdAndJurisdictionAndMagistrate(
                request.getId(), userMagistrate.getJurisdiction(), userMagistrate))
        .willReturn(Optional.of(request));

    // WHEN
    Request requestInvolved = courtService.getRequest(request.getId(), userMagistrate);

    // THEN
    then(requestInvolved.getId()).isEqualTo(request.getId());
  }

  @Test
  public void getRequest_not_involved() {

    // GIVEN
    Request request = makeRequest(RequestStatus.OPEN, false).build();
    UserMagistrate userMagistrate = makeMagistrate();
    // the user is not in the request's jurisdiction anymore
    userMagistrate.setJurisdiction(new Jurisdiction()); // no impact but just to explain the context
    given(
            requestRepository.findByIdAndJurisdictionAndMagistrate(
                request.getId(), userMagistrate.getJurisdiction(), userMagistrate))
        .willReturn(Optional.empty());

    // WHEN
    Throwable throwable =
        catchThrowable(() -> courtService.getRequest(request.getId(), userMagistrate));

    // THEN
    then(throwable)
        .isInstanceOf(AppException.class)
        .hasMessageContaining("magistrate not involved in this request");
  }
}
