package fr.gouv.beta.selexpert.service;

import static org.assertj.core.api.BDDAssertions.catchThrowable;
import static org.assertj.core.api.BDDAssertions.then;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.BDDMockito.given;
import static org.mockito.Mockito.never;
import static org.mockito.Mockito.verify;

import fr.gouv.beta.selexpert.AppException;
import fr.gouv.beta.selexpert.dto.AddNote;
import fr.gouv.beta.selexpert.dto.ExpertUpdateData.UpdatableData;
import fr.gouv.beta.selexpert.dto.RequestRefusalForm;
import fr.gouv.beta.selexpert.event.ExpertIsNowAvailableEvent;
import fr.gouv.beta.selexpert.event.ExpertIsStillUnavailableEvent;
import fr.gouv.beta.selexpert.event.QuestionAskedEvent;
import fr.gouv.beta.selexpert.event.RequestAcceptedEvent;
import fr.gouv.beta.selexpert.event.RequestRefusedAutoEvent;
import fr.gouv.beta.selexpert.event.RequestRefusedEvent;
import fr.gouv.beta.selexpert.model.Expert;
import fr.gouv.beta.selexpert.model.Note;
import fr.gouv.beta.selexpert.model.NoteType;
import fr.gouv.beta.selexpert.model.RefusalReason;
import fr.gouv.beta.selexpert.model.Request;
import fr.gouv.beta.selexpert.model.Request.RequestBuilder;
import fr.gouv.beta.selexpert.model.RequestExpert;
import fr.gouv.beta.selexpert.model.RequestExpert.RequestExpertBuilder;
import fr.gouv.beta.selexpert.model.RequestExpertId;
import fr.gouv.beta.selexpert.model.RequestExpertStatus;
import fr.gouv.beta.selexpert.model.RequestStatus;
import fr.gouv.beta.selexpert.model.UserExpert;
import fr.gouv.beta.selexpert.repository.ExpertRepository;
import fr.gouv.beta.selexpert.repository.NoteRepository;
import fr.gouv.beta.selexpert.repository.RequestExpertRepository;
import fr.gouv.beta.selexpert.repository.UserExpertRepository;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.ZoneId;
import java.time.format.DateTimeFormatter;
import java.util.ArrayList;
import java.util.List;
import java.util.Optional;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.extension.ExtendWith;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.EnumSource;
import org.junit.jupiter.params.provider.EnumSource.Mode;
import org.junit.jupiter.params.provider.ValueSource;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.junit.jupiter.MockitoExtension;
import org.springframework.context.ApplicationEventPublisher;
import org.springframework.test.util.ReflectionTestUtils;

@ExtendWith(MockitoExtension.class)
class ExpertServiceTest {

  @Mock private RequestExpertRepository requestExpertRepository;
  @Mock private UserExpertRepository userExpertRepository;
  @Mock private ExpertRepository expertRepository;
  @Mock private NoteRepository noteRepository;
  @Mock private ApplicationEventPublisher eventPublisher;

  @InjectMocks private ExpertService expertService;

  private static RequestExpertBuilder makeRequestExpert(
      RequestExpertStatus status, Request request, Integer expertId) {
    Expert expert = new Expert();
    expert.setId(expertId);
    return RequestExpert.builder().status(status).request(request).expert(expert);
  }

  private static RequestBuilder makeRequest() {
    return Request.builder().id(1).status(RequestStatus.OPEN).requestExperts(new ArrayList<>());
  }

  @BeforeEach
  public void setUp() {
    ReflectionTestUtils.setField(
        expertService, "dtfUnavailable", DateTimeFormatter.ofPattern("dd/MM/yy"));
    ReflectionTestUtils.setField(expertService, "weekdaysToReply", 2);
  }

  @Test
  public void acceptRequest() {

    // GIVEN
    Request request = makeRequest().build();
    RequestExpert requestExpert = makeRequestExpert(RequestExpertStatus.SENT, request, 1).build();
    RequestExpert requestExpert2 =
        makeRequestExpert(RequestExpertStatus.QUEUED, request, 2).build();
    request.getRequestExperts().addAll(List.of(requestExpert, requestExpert2));
    given(requestExpertRepository.findById(new RequestExpertId(1, 1)))
        .willReturn(Optional.of(requestExpert));

    // WHEN
    expertService.acceptRequest(1, 1);

    // THEN
    then(requestExpert.getStatus()).isEqualTo(RequestExpertStatus.ACCEPTED);
    then(requestExpert2.getStatus()).isEqualTo(RequestExpertStatus.FULFILLED);
    then(request.getStatus()).isEqualTo(RequestStatus.ACCEPTED);
    verify(eventPublisher)
        .publishEvent(new RequestAcceptedEvent(request, requestExpert.getExpert()));
  }

  @ParameterizedTest
  @ValueSource(
      strings = {"requestExpertNotFound", "requestExpertStatusNotSent", "requestStatusNotOpen"})
  public void acceptRequest_cant_accept(String scenario) {

    // GIVEN
    Request request = makeRequest().build();
    if (scenario.equals("requestStatusNotOpen")) {
      request.setStatus(RequestStatus.ACCEPTED);
    }
    RequestExpert requestExpert =
        makeRequestExpert(
                scenario.equals("requestExpertStatusNotSent")
                    ? RequestExpertStatus.QUEUED
                    : RequestExpertStatus.SENT,
                request,
                1)
            .build();
    request.getRequestExperts().add(requestExpert);
    given(requestExpertRepository.findById(new RequestExpertId(1, 1)))
        .willReturn(
            scenario.equals("requestExpertNotFound")
                ? Optional.empty()
                : Optional.of(requestExpert));

    // WHEN
    Throwable throwable = catchThrowable(() -> expertService.acceptRequest(1, 1));

    // THEN
    then(throwable)
        .isInstanceOf(AppException.class)
        .hasMessageContaining(
            scenario.equals("requestExpertNotFound")
                ? "expert not involved in this request"
                : "can't act on this request");
  }

  @ParameterizedTest
  @ValueSource(ints = {1, 2})
  public void refuseRequest(int nbRequestExperts) {

    // GIVEN
    RequestRefusalForm requestRefusalForm =
        RequestRefusalForm.builder()
            .reason(RefusalReason.UNAVAILABLE)
            .unavailableUntil(LocalDate.parse(nbRequestExperts + "999-12-31"))
            .build();
    UserExpert userExpert = new UserExpert();
    Request request = makeRequest().build();
    RequestExpert requestExpert = makeRequestExpert(RequestExpertStatus.SENT, request, 1).build();
    requestExpert.getExpert().setUserExpert(userExpert);
    RequestExpert requestExpert2 = null;
    if (nbRequestExperts == 2) {
      requestExpert2 = makeRequestExpert(RequestExpertStatus.QUEUED, request, 2).build();
      request.getRequestExperts().add(requestExpert2);
    }
    given(requestExpertRepository.findById(new RequestExpertId(1, 1)))
        .willReturn(Optional.of(requestExpert));
    given(userExpertRepository.findByExpertId(1)).willReturn(Optional.of(userExpert));

    // WHEN
    expertService.refuseRequest(1, 1, requestRefusalForm);

    // THEN
    then(requestExpert.getStatus()).isEqualTo(RequestExpertStatus.REFUSED);
    if (nbRequestExperts == 1) {
      then(request.getStatus()).isEqualTo(RequestStatus.REFUSED);
      then(userExpert.getUnavailableUntil()).isNull(); // because 1999
      then(requestExpert.getRefusalJustification()).isEqualTo("pour une durée indéterminée");
    }
    if (nbRequestExperts == 2) {
      then(request.getStatus()).isEqualTo(RequestStatus.OPEN);
      then(requestExpert2.getStatus()).isEqualTo(RequestExpertStatus.SENDING);
      then(userExpert.getUnavailableUntil()).isNotNull(); // because 2999
      then(requestExpert.getRefusalJustification()).isEqualTo("jusqu'au 31/12/99");
    }
    then(userExpert.getUnavailableFrom()).isNotNull();
    verify(eventPublisher)
        .publishEvent(
            new RequestRefusedEvent(
                request,
                requestExpert.getExpert(),
                requestRefusalForm.getReason(),
                requestRefusalForm.getJustification()));
  }

  @ParameterizedTest
  @EnumSource(
      value = RefusalReason.class,
      names = {"GEO_CONSTRAINT", "OTHER"})
  public void refuseRequest_error(RefusalReason refusalReason) {

    // GIVEN
    RequestRefusalForm requestRefusalForm =
        RequestRefusalForm.builder().reason(refusalReason).build();
    UserExpert userExpert = new UserExpert();
    Request request = makeRequest().build();
    RequestExpert requestExpert = makeRequestExpert(RequestExpertStatus.SENT, request, 1).build();
    requestExpert.getExpert().setUserExpert(userExpert);
    given(requestExpertRepository.findById(new RequestExpertId(1, 1)))
        .willReturn(Optional.of(requestExpert));
    given(userExpertRepository.findByExpertId(1)).willReturn(Optional.of(userExpert));

    // WHEN
    Throwable throwable =
        catchThrowable(() -> expertService.refuseRequest(1, 1, requestRefusalForm));

    // THEN
    then(throwable)
        .isInstanceOf(AppException.class)
        .hasMessageContaining("can't use " + refusalReason);
  }

  @Test
  public void refuseRequestAuto() {

    // GIVEN
    var now = LocalDateTime.now(ZoneId.of("Europe/Paris"));
    Request request = makeRequest().build();
    RequestExpert requestExpert1 =
        makeRequestExpert(RequestExpertStatus.REFUSED, request, 1).build();
    RequestExpert requestExpert2 = makeRequestExpert(RequestExpertStatus.SENT, request, 2).build();
    requestExpert2.setSentOn(now.minusDays(10));
    UserExpert userExpert = new UserExpert();
    requestExpert2.getExpert().setUserExpert(userExpert);
    RequestExpert requestExpert3 =
        makeRequestExpert(RequestExpertStatus.QUEUED, request, 3).build();
    request.setRequestExperts(List.of(requestExpert1, requestExpert2, requestExpert3));
    given(requestExpertRepository.findByStatusIn(RequestExpertStatus.SENT))
        .willReturn(List.of(requestExpert2));

    // WHEN
    expertService.refuseRequestAuto();

    // THEN
    then(requestExpert1.getStatus()).isEqualTo(RequestExpertStatus.REFUSED);
    then(requestExpert2.getStatus()).isEqualTo(RequestExpertStatus.REFUSED_AUTO);
    then(requestExpert2.getExpert().getUserExpert().getUnavailableFrom())
        .isEqualTo(now.toLocalDate());
    then(requestExpert3.getStatus()).isEqualTo(RequestExpertStatus.SENDING);
    verify(eventPublisher)
        .publishEvent(new RequestRefusedAutoEvent(request, List.of(requestExpert2.getExpert())));
  }

  @ParameterizedTest
  @EnumSource(value = UpdatableData.class)
  public void updateExpertData(UpdatableData field) {

    // GIVEN
    Expert expert = new Expert();
    UserExpert userExpert = new UserExpert();
    expert.setUserExpert(userExpert);
    given(expertRepository.findById(1)).willReturn(Optional.of(expert));

    // WHEN
    String cleanValue =
        expertService.updateExpertData(
            1, field, field != UpdatableData.opalexe ? "\n\n\na\n\n\nb  " : "true");

    // THEN
    String updatedField =
        switch (field) {
          case phoneMobile -> expert.getPhoneMobile();
          case phoneLandline -> expert.getPhoneLandline();
          case specialty -> expert.getSpecialty();
          case address -> expert.getAddress();
          case city -> expert.getCity();
          case zipCode -> expert.getZipCode();
          case geoConstraint -> expert.getUserExpert().getGeoConstraint();
          case email -> expert.getEmail();
          case name -> expert.getName();
          case opalexe -> expert.getUserExpert().getOpalexe().toString();
        };
    then(updatedField).isEqualTo(cleanValue);
    if (field == UpdatableData.email) then(expert.getUserExpert().getEmail()).isEqualTo(cleanValue);
    if (field == UpdatableData.name) then(expert.getUserExpert().getName()).isEqualTo(cleanValue);
  }

  @Test
  public void updateExpertData_geoConstraint_noUserExpert() {

    // GIVEN
    given(expertRepository.findById(1)).willReturn(Optional.of(new Expert()));

    // WHEN
    Throwable throwable =
        catchThrowable(() -> expertService.updateExpertData(1, UpdatableData.geoConstraint, ""));

    // THEN
    then(throwable)
        .isInstanceOf(AppException.class)
        .hasMessageContaining("can't update geoConstraint");
  }

  @Test
  public void checkAvailability_expertIsNowAvailable() {

    // GIVEN
    LocalDate now = LocalDate.now(ZoneId.of("Europe/Paris"));
    UserExpert userExpert = new UserExpert();
    userExpert.setUnavailableFrom(now.minusDays(60));
    userExpert.setUnavailableUntil(now.minusDays(1));
    Expert expert = new Expert();
    userExpert.setExpert(expert);
    given(userExpertRepository.findByUnavailableFromIsNotNullOrUnavailableUntilIsNotNull())
        .willReturn(List.of(userExpert));

    // WHEN
    expertService.checkAvailability();

    // THEN
    then(userExpert.getUnavailableFrom()).isNull();
    then(userExpert.getUnavailableUntil()).isNull();
    verify(eventPublisher).publishEvent(new ExpertIsNowAvailableEvent(expert));
  }

  @ParameterizedTest
  @ValueSource(ints = {7, 14, 30, 60, 90})
  public void checkAvailability_expertIsStillUnavailable(int daysSince) {

    // GIVEN
    LocalDate now = LocalDate.now(ZoneId.of("Europe/Paris"));
    UserExpert userExpert = new UserExpert();
    userExpert.setUnavailableFrom(now.minusDays(daysSince));
    Expert expert = new Expert();
    userExpert.setExpert(expert);
    given(userExpertRepository.findByUnavailableFromIsNotNullOrUnavailableUntilIsNotNull())
        .willReturn(List.of(userExpert));

    // WHEN
    expertService.checkAvailability();

    // THEN
    then(userExpert.getUnavailableFrom()).isNotNull();
    then(userExpert.getUnavailableUntil()).isNull();
    verify(eventPublisher).publishEvent(new ExpertIsStillUnavailableEvent(expert));
  }

  @Test
  public void checkAvailability_deletedExpert() {

    // GIVEN
    LocalDate now = LocalDate.now(ZoneId.of("Europe/Paris"));
    UserExpert userExpert = new UserExpert();
    userExpert.setUnavailableFrom(now.minusDays(60));
    Expert expert = new Expert();
    expert.setDeleted(true);
    userExpert.setExpert(expert);
    given(userExpertRepository.findByUnavailableFromIsNotNullOrUnavailableUntilIsNotNull())
        .willReturn(List.of(userExpert));

    // WHEN
    expertService.checkAvailability();

    // THEN
    then(userExpert.getUnavailableFrom()).isEqualTo(now.minusDays(60));
    verify(eventPublisher, never()).publishEvent(new ExpertIsNowAvailableEvent(expert));
  }

  @ParameterizedTest
  @EnumSource(
      value = RequestExpertStatus.class,
      names = {"SENDING", "SENT", "ACCEPTED"},
      mode = Mode.INCLUDE)
  public void addQuestion(RequestExpertStatus status) {

    // GIVEN
    AddNote addNote = AddNote.builder().requestId(1).content("hello").build();
    UserExpert userExpert = new UserExpert();
    Request request = makeRequest().build();
    RequestExpert requestExpert = makeRequestExpert(status, request, 1).build();
    requestExpert.getExpert().setUserExpert(userExpert);
    userExpert.setExpert(requestExpert.getExpert());
    given(requestExpertRepository.findById(new RequestExpertId(1, 1)))
        .willReturn(Optional.of(requestExpert));

    // WHEN
    Note question = expertService.addQuestion(addNote, userExpert);

    // THEN
    verify(noteRepository).save(any());
    verify(eventPublisher).publishEvent(new QuestionAskedEvent(request, requestExpert.getExpert()));
    then(question.getContent()).isEqualTo(addNote.getContent());
    then(question.getType()).isEqualTo(NoteType.QUESTION);
    then(question.getRequest()).isEqualTo(request);
    then(question.getUser()).isEqualTo(userExpert);
  }

  @ParameterizedTest
  @EnumSource(
      value = RequestExpertStatus.class,
      names = {"SENDING", "SENT", "ACCEPTED"},
      mode = Mode.EXCLUDE)
  public void addQuestion_status_not_allowed(RequestExpertStatus status) {

    // GIVEN
    AddNote addNote = AddNote.builder().requestId(1).content("hello").build();
    UserExpert userExpert = new UserExpert();
    Request request = makeRequest().build();
    RequestExpert requestExpert = makeRequestExpert(status, request, 1).build();
    requestExpert.getExpert().setUserExpert(userExpert);
    userExpert.setExpert(requestExpert.getExpert());
    given(requestExpertRepository.findById(new RequestExpertId(1, 1)))
        .willReturn(Optional.of(requestExpert));

    // WHEN
    Throwable throwable = catchThrowable(() -> expertService.addQuestion(addNote, userExpert));

    // THEN
    then(throwable)
        .isInstanceOf(AppException.class)
        .hasMessageContaining("can't act on this request");
  }
}
