package fr.gouv.beta.selexpert.service;

import static org.assertj.core.api.BDDAssertions.then;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.BDDMockito.given;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;

import fr.gouv.beta.selexpert.event.ClerksAddedEvent;
import fr.gouv.beta.selexpert.event.RequestAcceptedEvent;
import fr.gouv.beta.selexpert.event.RequestCreatedEvent;
import fr.gouv.beta.selexpert.event.RequestRefusedAutoEvent;
import fr.gouv.beta.selexpert.event.RequestRefusedEvent;
import fr.gouv.beta.selexpert.model.Expert;
import fr.gouv.beta.selexpert.model.Jurisdiction;
import fr.gouv.beta.selexpert.model.RefusalReason;
import fr.gouv.beta.selexpert.model.Request;
import fr.gouv.beta.selexpert.model.Request.RequestBuilder;
import fr.gouv.beta.selexpert.model.RequestExpert;
import fr.gouv.beta.selexpert.model.RequestExpert.RequestExpertBuilder;
import fr.gouv.beta.selexpert.model.RequestExpertStatus;
import fr.gouv.beta.selexpert.model.RequestStatus;
import fr.gouv.beta.selexpert.model.UserClerk;
import fr.gouv.beta.selexpert.model.UserExpert;
import fr.gouv.beta.selexpert.model.UserMagistrate;
import fr.gouv.beta.selexpert.repository.RequestExpertRepository;
import fr.gouv.beta.selexpert.service.MailService.Email;
import freemarker.template.Configuration;
import freemarker.template.Template;
import java.io.IOException;
import java.time.format.DateTimeFormatter;
import java.util.ArrayList;
import java.util.List;
import java.util.Locale;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.extension.ExtendWith;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.EnumSource;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.junit.jupiter.MockitoExtension;
import org.springframework.test.util.ReflectionTestUtils;
import sendinblue.ApiException;

@ExtendWith(MockitoExtension.class)
class MailServiceTest {

  @Mock private Configuration freeMarkerConfiguration;
  @Mock private RequestExpertRepository requestExpertRepository;
  @Mock private MailSenderService mailSenderService;

  @InjectMocks private MailService mailService;

  private static RequestExpertBuilder makeRequestExpert(
      RequestExpertStatus status, Request request, Integer expertId, boolean hasUserExpert) {
    Expert expert = new Expert();
    expert.setId(expertId);
    expert.setEmail(expertId + "@expert.com");
    if (hasUserExpert)
      expert.setUserExpert(UserExpert.builder().email(expertId + "@userexpert.com").build());
    return RequestExpert.builder().status(status).request(request).expert(expert);
  }

  private static RequestBuilder makeRequest() {
    return Request.builder()
        .id(1)
        .jurisdiction(Jurisdiction.builder().name("").build())
        .status(RequestStatus.OPEN)
        .requestExperts(new ArrayList<>())
        .magistrate(
            UserMagistrate.builder()
                .email("1@magistrate.com")
                .jurisdiction(new Jurisdiction())
                .build());
  }

  private static UserClerk makeClerk(Integer clerkId, Boolean allNotifs) {
    return UserClerk.builder()
        .id(clerkId)
        .email(clerkId + "@clerk.com")
        .allNotifications(allNotifs)
        .build();
  }

  @BeforeEach
  public void setUp() throws IOException {
    ReflectionTestUtils.setField(mailService, "senderEmail", "contact@selexpert.beta.gouv.fr");
    ReflectionTestUtils.setField(
        mailService, "activationUrl", "https://selexpert/activation-compte/");
    ReflectionTestUtils.setField(
        mailService, "passwordResetUrl", "https://selexpert/mot-de-passe/");
    ReflectionTestUtils.setField(mailService, "selexpertUrl", "https://selexpert/");
    ReflectionTestUtils.setField(mailService, "isProduction", true);
    ReflectionTestUtils.setField(mailService, "weekdaysToReply", 2);
    ReflectionTestUtils.setField(
        mailService,
        "dtfDeadline",
        DateTimeFormatter.ofPattern("EEEE dd MMMM yyyy à HH:mm:ss").withLocale(Locale.FRANCE));
    given(freeMarkerConfiguration.getTemplate(any()))
        .willReturn(Template.getPlainTextTemplate("", "", null));
  }

  @Test
  public void onRequestCreated_send_to_experts() throws ApiException {

    // GIVEN
    // first expert is not activated = send email to Expert's email + email to selexpert team
    // second expert is activated = send email to User's email
    Request request = makeRequest().build();
    RequestExpert requestExpert =
        makeRequestExpert(RequestExpertStatus.SENDING, request, 1, false).build();
    RequestExpert requestExpert2 =
        makeRequestExpert(RequestExpertStatus.SENDING, request, 2, true).build();
    request.getRequestExperts().addAll(List.of(requestExpert, requestExpert2));
    given(requestExpertRepository.findByRequestAndStatus(request, RequestExpertStatus.SENDING))
        .willReturn(List.of(requestExpert, requestExpert2));

    // WHEN
    mailService.onRequestCreated(new RequestCreatedEvent(request));

    // THEN
    then(requestExpert.getStatus()).isEqualTo(RequestExpertStatus.SENT);
    then(requestExpert2.getStatus()).isEqualTo(RequestExpertStatus.SENT);
    then(requestExpert.getSentOn()).isNotNull();
    then(requestExpert.getSentOn()).isEqualTo(requestExpert2.getSentOn());
    verify(mailSenderService).send("1@expert.com", Email.REQUEST_CREATED.getSubject(), "");
    verify(mailSenderService).send("2@userexpert.com", Email.REQUEST_CREATED.getSubject(), "");
    verify(mailSenderService)
        .send("contact@selexpert.beta.gouv.fr", Email.EXPERT_NOT_ACTIVATED.getSubject(), "");
  }

  @Test
  public void onRequestCreated_send_to_clerks() throws ApiException {

    // GIVEN
    Request request = makeRequest().build();
    UserClerk clerkCreator = makeClerk(1, null);
    UserClerk clerkAssigned = makeClerk(2, null);
    request.setClerk(clerkCreator);
    request.setRequestClerks(List.of(clerkCreator, clerkAssigned));
    given(requestExpertRepository.findByRequestAndStatus(request, RequestExpertStatus.SENDING))
        .willReturn(List.of());

    // WHEN
    mailService.onRequestCreated(new RequestCreatedEvent(request));

    // THEN
    verify(mailSenderService, times(0))
        .send("1@clerk.com", Email.REQUEST_ASSIGNED.getSubject(), "");
    verify(mailSenderService, times(1))
        .send("2@clerk.com", Email.REQUEST_ASSIGNED.getSubject(), "");
  }

  @Test
  public void onClerksAdded() throws ApiException {

    // GIVEN
    UserClerk clerkAllNotifsNull = makeClerk(1, null);
    UserClerk clerkAllNotifsTrue = makeClerk(2, true);
    UserClerk clerkAllNotifsFalse = makeClerk(3, false);

    // WHEN
    mailService.onClerksAdded(
        new ClerksAddedEvent(
            new Request(), List.of(clerkAllNotifsNull, clerkAllNotifsTrue, clerkAllNotifsFalse)));

    // THEN
    verify(mailSenderService, times(1))
        .send("1@clerk.com", Email.REQUEST_ASSIGNED.getSubject(), "");
    verify(mailSenderService, times(1))
        .send("2@clerk.com", Email.REQUEST_ASSIGNED.getSubject(), "");
    verify(mailSenderService, times(0))
        .send("3@clerk.com", Email.REQUEST_ASSIGNED.getSubject(), "");
  }

  @Test
  public void onRequestAccepted() throws ApiException {

    // GIVEN
    Request request = makeRequest().build();
    request.setUrgent(true);
    RequestExpert requestExpertAccepted =
        makeRequestExpert(RequestExpertStatus.ACCEPTED, request, 1, true).build();
    RequestExpert requestExpertFulfilled =
        makeRequestExpert(RequestExpertStatus.FULFILLED, request, 2, false).build();
    request.getRequestExperts().addAll(List.of(requestExpertAccepted, requestExpertFulfilled));
    request.setRequestClerks(List.of(makeClerk(1, true), makeClerk(2, false)));

    // WHEN
    mailService.onRequestAccepted(
        new RequestAcceptedEvent(request, requestExpertAccepted.getExpert()));

    // THEN
    verify(mailSenderService).send("1@magistrate.com", Email.REQUEST_ACCEPTED.getSubject(), "");
    verify(mailSenderService).send("1@clerk.com", Email.REQUEST_ACCEPTED.getSubject(), "");
    verify(mailSenderService).send("2@clerk.com", Email.REQUEST_ACCEPTED.getSubject(), "");
    verify(mailSenderService).send("1@userexpert.com", Email.REQUEST_ACCEPTED_ACK.getSubject(), "");
    verify(mailSenderService).send("2@expert.com", Email.REQUEST_FULFILLED.getSubject(), "");
  }

  @ParameterizedTest
  @EnumSource(
      value = RequestStatus.class,
      names = {"OPEN", "REFUSED"})
  public void onRequestRefused(RequestStatus requestStatus) throws ApiException {

    // GIVEN
    boolean open = requestStatus.equals(RequestStatus.OPEN);
    Request request = makeRequest().build();
    request.setStatus(requestStatus);
    RequestExpert requestExpertRefused =
        makeRequestExpert(RequestExpertStatus.REFUSED, request, 1, true).build();
    request.getRequestExperts().add(requestExpertRefused);
    if (open) {
      RequestExpert requestExpertSending =
          makeRequestExpert(RequestExpertStatus.SENDING, request, 2, true).build();
      request.getRequestExperts().add(requestExpertSending);
      given(requestExpertRepository.findByRequestAndStatus(request, RequestExpertStatus.SENDING))
          .willReturn(List.of(requestExpertSending));
    }

    // WHEN
    mailService.onRequestRefused(
        new RequestRefusedEvent(
            request, requestExpertRefused.getExpert(), RefusalReason.OTHER, null));

    // THEN
    verify(mailSenderService)
        .send(
            "1@magistrate.com",
            open ? Email.REQUEST_REFUSED.getSubject() : Email.REQUEST_REFUSED_ALL.getSubject(),
            "");
    verify(mailSenderService).send("1@userexpert.com", Email.REQUEST_REFUSED_ACK.getSubject(), "");
    verify(mailSenderService, times(open ? 1 : 0))
        .send("2@userexpert.com", Email.REQUEST_CREATED.getSubject(), "");
  }

  @Test
  public void onRequestRefusedAuto() throws ApiException {

    // GIVEN
    Request request = makeRequest().build();
    request.setStatus(RequestStatus.REFUSED);
    RequestExpert requestExpertAuto =
        makeRequestExpert(RequestExpertStatus.REFUSED_AUTO, request, 1, false).build();
    request.getRequestExperts().add(requestExpertAuto);

    // WHEN
    mailService.onRequestRefusedAuto(
        new RequestRefusedAutoEvent(request, List.of(requestExpertAuto.getExpert())));

    // THEN
    verify(mailSenderService).send("1@magistrate.com", Email.REQUEST_REFUSED_ALL.getSubject(), "");
    verify(mailSenderService).send("1@expert.com", Email.REQUEST_REFUSED_AUTO.getSubject(), "");
  }
}
